import numpy as np
import cv2
import pathlib

from . import data_files, drawing, ocv


class Gaze:
    # description of tsv file used for storage
    _columns_compressed = {'timestamp': 1, 'timestamp_VOR': 1, 'timestamp_ref': 1, 'frame_idx':1, 'frame_idx_VOR':1, 'frame_idx_ref':1,
                           'gaze_pos_vid': 2, 'gaze_pos_3d': 3, 'gaze_dir_l': 3, 'gaze_ori_l': 3, 'gaze_dir_r': 3, 'gaze_ori_r': 3}
    _non_float          = {'frame_idx': int, 'frame_idx_VOR': int, 'frame_idx_ref': int}
    _columns_optional   = ['timestamp_VOR', 'frame_idx_VOR', 'timestamp_ref', 'frame_idx_ref']

    def __init__(self,
                 timestamp      : float,
                 frame_idx      : int,

                 gaze_pos_vid   : np.ndarray,

                 timestamp_ori  : float      = None,
                 frame_idx_ori  : int        = None,
                 timestamp_VOR  : float      = None,
                 frame_idx_VOR  : int        = None,
                 timestamp_ref  : float      = None,
                 frame_idx_ref  : int        = None,

                 gaze_pos_3d    : np.ndarray = None,
                 gaze_dir_l     : np.ndarray = None,
                 gaze_ori_l     : np.ndarray = None,
                 gaze_dir_r     : np.ndarray = None,
                 gaze_ori_r     : np.ndarray = None):
        self.timestamp    : float       = timestamp
        self.frame_idx    : int         = frame_idx

        # optional timestamps and frame_idxs
        self.timestamp_ori: float       = timestamp_ori     # timestamp field can be set to any of these three. Keep here a copy of the original timestamp
        self.frame_idx_ori: int         = frame_idx_ori
        self.timestamp_VOR: float       = timestamp_VOR
        self.frame_idx_VOR: int         = frame_idx_VOR
        self.timestamp_ref: float       = timestamp_ref
        self.frame_idx_ref: int         = frame_idx_ref     # frameidx _in reference video_ not in this eye tracker's video (unless this is the reference)

        self.gaze_pos_vid : np.ndarray  = gaze_pos_vid      # gaze point on the scene video
        self.gaze_pos_3d  : np.ndarray  = gaze_pos_3d       # gaze point in the world (often binocular gaze point)
        self.gaze_dir_l   : np.ndarray  = gaze_dir_l
        self.gaze_ori_l   : np.ndarray  = gaze_ori_l
        self.gaze_dir_r   : np.ndarray  = gaze_dir_r
        self.gaze_ori_r   : np.ndarray  = gaze_ori_r

    def draw(self, img:np.ndarray, camera_params:ocv.CameraParams=None, sub_pixel_fac=1, clr=(0,255,0), draw_3d_gaze_point=True):
        drawing.openCVCircle(img, self.gaze_pos_vid, 8, clr, 2, sub_pixel_fac)
        # draw 3D gaze point as well, usually coincides with 2D gaze point, but not always. E.g. the Adhawk MindLink may
        # apply a correction for parallax error to the projected gaze point using the vergence signal.
        if draw_3d_gaze_point and self.gaze_pos_3d is not None and camera_params.has_intrinsics():
            camRot = np.zeros((1,3)) if camera_params.rotation_vec is None else camera_params.rotation_vec
            camPos = np.zeros((1,3)) if camera_params.position     is None else camera_params.position
            a = cv2.projectPoints(np.array(self.gaze_pos_3d).reshape(1,3),camRot,camPos,camera_params.camera_mtx,camera_params.distort_coeffs)[0][0][0]
            drawing.openCVCircle(img, a, 5, (0,255,255), -1, sub_pixel_fac)


def read_dict_from_file(file_name:str|pathlib.Path, episodes:list[list[int]]=None, ts_column_suffixes: list[str] = None) -> tuple[dict[int,list[Gaze]], int]:
    return data_files.read_file(file_name,
                                Gaze, False, False, True, True,
                                episodes=episodes, ts_fridx_field_suffixes=ts_column_suffixes)

def write_dict_to_file(gazes: list[Gaze] | dict[int,list[Gaze]], file_name:str|pathlib.Path, skip_missing=False):
    data_files.write_array_to_file(gazes, file_name,
                                   Gaze._columns_compressed,
                                   Gaze._columns_optional,
                                   skip_all_nan=skip_missing)


import random
import functools
import sys
import traceback
from typing import Any, Callable
from imgui_bundle import imgui, hello_imgui, icons_fontawesome_6 as ifa6
import glfw

from . import msg_box
from .. import platform, utils


# https://gist.github.com/Willy-JL/f733c960c6b0d2284bcbee0316f88878
def get_traceback(*exc_info: list):
    exc_info = exc_info or sys.exc_info()
    tb_lines = traceback.format_exception(*exc_info)
    tb = "".join(tb_lines)
    return tb


def draw_tooltip(hover_text):
    imgui.begin_tooltip()
    imgui.push_text_wrap_pos(min(imgui.get_font_size() * 35, imgui.get_io().display_size.x))
    imgui.text_unformatted(hover_text)
    imgui.pop_text_wrap_pos()
    imgui.end_tooltip()

def draw_hover_text(hover_text: str, text="(?)", force=False, hovered_flags=imgui.HoveredFlags_.for_tooltip|imgui.HoveredFlags_.delay_normal, *args, **kwargs):
    if text:
        imgui.text_disabled(text, *args, **kwargs)
    if force or imgui.is_item_hovered(hovered_flags):
        draw_tooltip(hover_text)
        return True
    return False


def rand_num_str(len=8):
    return "".join((random.choice('0123456789') for _ in range(len)))

def push_popup(gui, *args, bottom=False, **kwargs):
    if len(args) + len(kwargs) > 1:
        if args[0] in (popup, msg_box.msgbox):
            args = list(args)
            args[1] = args[1] + "##popup_" + rand_num_str()
        popup_func = functools.partial(*args, **kwargs)
    else:
        popup_func = args[0]
    if bottom:
        gui.popup_stack.insert(0, popup_func)
    else:
        gui.popup_stack.append(popup_func)
    return popup_func

def fix_popup_transparency():
    frame_bg_col = imgui.get_style().color_(imgui.Col_.title_bg_active)
    imgui.get_style().set_color_(imgui.Col_.title_bg_active,(frame_bg_col.x, frame_bg_col.y, frame_bg_col.z, 1.))
    popup_bg_col = imgui.get_style().color_(imgui.Col_.popup_bg)
    imgui.get_style().set_color_(imgui.Col_.popup_bg,(popup_bg_col.x, popup_bg_col.y, popup_bg_col.z, 1.))

def handle_popup_stack(popup_stack: list):
    fix_popup_transparency()
    open_popup_count = 0
    for popup in popup_stack:
        if hasattr(popup, "tick"):
            popup_func = popup.tick
        else:
            popup_func = popup
        opened, closed = popup_func()
        if closed:
            popup_stack.remove(popup)
        open_popup_count += opened
    # Popups are closed all at the end to allow stacking
    for _ in range(open_popup_count):
        imgui.end_popup()

def close_weak_popup(check_escape: bool = True, check_click_outside: bool = True):
    if not imgui.is_popup_open("", imgui.PopupFlags_.any_popup_id):
        # This is the topmost popup
        if check_escape and imgui.is_key_pressed(imgui.Key.escape):
            # Escape is pressed
            imgui.close_current_popup()
            return True
        elif check_click_outside and imgui.is_mouse_clicked(imgui.MouseButton_.left):
            # Mouse was just clicked
            pos = imgui.get_window_pos()
            size = imgui.get_window_size()
            if not imgui.is_mouse_hovering_rect(pos, (pos.x+size.x, pos.y+size.y), clip=False):
                # Popup is not hovered
                imgui.close_current_popup()
                return True
    return False

popup_flags: int = (
    imgui.WindowFlags_.no_collapse |
    imgui.WindowFlags_.no_saved_settings |
    imgui.WindowFlags_.always_auto_resize
)

def popup(label: str, popup_content: Callable, buttons: dict[str, Callable] = None, closable=True, escape=True, outside=True):
    if buttons is True:
        buttons = {
            ifa6.ICON_FA_CHECK + " Ok": None
        }
    if not imgui.is_popup_open(label):
        imgui.open_popup(label)
    closed = False
    opened = 1
    if imgui.begin_popup_modal(label, closable or None, flags=popup_flags)[0]:
        if escape or outside:
            closed = close_weak_popup(check_escape=escape, check_click_outside=outside)
        imgui.begin_group()
        activate_button = popup_content()
        imgui.end_group()
        imgui.spacing()
        if buttons:
            btns_width = sum(imgui.calc_text_size(name).x for name in buttons) + (2 * len(buttons) * imgui.get_style().frame_padding.x) + (imgui.get_style().item_spacing.x * (len(buttons) - 1))
            cur_pos_x = imgui.get_cursor_pos_x()
            new_pos_x = cur_pos_x + imgui.get_content_region_avail().x - btns_width
            if new_pos_x > cur_pos_x:
                imgui.set_cursor_pos_x(new_pos_x)
            for i, (label,callbacks) in enumerate(buttons.items()):
                if isinstance(callbacks,tuple):
                    callback = callbacks[0]
                    disabled = len(callbacks)>1 and callbacks[1]()
                else:
                    callback = callbacks
                    disabled = False
                if disabled:
                    imgui.begin_disabled()
                if imgui.button(label) or activate_button==i:
                    if callback:
                        callback()
                    imgui.close_current_popup()
                    closed = True
                if disabled:
                    imgui.end_disabled()
                imgui.same_line()
    else:
        opened = 0
        closed = True
    return opened, closed


def selectable_item_logic(iid: int|str, selected: dict[int,Any], last_clicked_id: int, sorted_ids: list[int],
                          selectable_clicked: bool, new_selectable_state: bool,
                          allow_multiple=True, overlayed_hovered=False, overlayed_clicked=False, new_overlayed_state=False):
    if overlayed_clicked:
        if not allow_multiple:
            utils.set_all(selected, False)
        selected[iid] = new_overlayed_state
        last_clicked_id = iid
    elif selectable_clicked and not overlayed_hovered: # don't enter this branch if interaction is with another overlaid actionable item
        if not allow_multiple:
            utils.set_all(selected, False)
            selected[iid] = new_selectable_state
        else:
            num_selected = sum([selected[id] for id in sorted_ids])
            if not imgui.get_io().key_ctrl:
                # deselect all, below we'll either select all, or range between last and current clicked
                utils.set_all(selected, False)

            if imgui.get_io().key_shift:
                # select range between last clicked and just clicked item
                idx              = sorted_ids.index(iid)
                last_clicked_idx = sorted_ids.index(last_clicked_id)
                idxs = sorted([idx, last_clicked_idx])
                for rid in range(idxs[0],idxs[1]+1):
                    selected[sorted_ids[rid]] = True
            else:
                selected[iid] = True if num_selected>1 and not imgui.get_io().key_ctrl else new_selectable_state

            # consistent with Windows behavior, only update last clicked when shift not pressed
            if not imgui.get_io().key_shift:
                last_clicked_id = iid

    return last_clicked_id

def handle_item_hitbox_events(iid: int, selected: dict[int,bool], context_menu: Callable[[int],bool]) -> tuple[bool,bool]:
    right_clicked = False
    selectables_edited = False
    # Right click = context menu
    if context_menu and imgui.begin_popup_context_item(f"##{iid}_context"):
        right_clicked = True
        selectables_edited = context_menu(iid)
        imgui.end_popup()
    else:
        right_clicked = imgui.is_item_clicked(imgui.MouseButton_.right)

    if right_clicked:
        # update selected items. same logic as windows explorer:
        # 1. if right-clicked on one of the selected items, regardless of what modifier is pressed, keep selection as is
        # 2. if right-clicked elsewhere than on one of the selected items:
        # 2a. if control is down pop up right-click menu for the selected items.
        # 2b. if control not down, deselect everything except clicked item (if any)
        # NB: popup not shown when shift or control are down, do not know why...
        if not selected[iid] and not imgui.get_io().key_ctrl:
            utils.set_all(selected, False)
            selected[iid] = True
    return right_clicked, selectables_edited


def my_checkbox(label: str, state: bool, frame_size: tuple=None, frame_padding_override: list=None, do_vertical_align=True):
    style = imgui.get_style()
    if state:
        imgui.push_style_color(imgui.Col_.frame_bg_hovered, style.color_(imgui.Col_.button_hovered))
        imgui.push_style_color(imgui.Col_.frame_bg, style.color_(imgui.Col_.button_hovered))
        imgui.push_style_color(imgui.Col_.check_mark, style.color_(imgui.Col_.text))
    if frame_size is not None:
        frame_padding = frame_padding_override if frame_padding_override else [style.frame_padding.x, style.frame_padding.y]
        imgui.push_style_var(imgui.StyleVar_.frame_padding, frame_size)
        imgui.push_style_var(imgui.StyleVar_.item_spacing, (0.,0.))
        imgui.begin_group()
        if do_vertical_align and frame_padding[1]:
            imgui.dummy((0,frame_padding[1]))
        imgui.dummy((frame_padding[0],0))
        imgui.same_line()
    result = imgui.checkbox(label, state)
    if frame_size is not None:
        imgui.end_group()
        imgui.pop_style_var(2)
    if state:
        imgui.pop_style_color(3)
    return result

def my_combo(*args, **kwargs):
    imgui.push_style_color(imgui.Col_.button, imgui.get_style().color_(imgui.Col_.button_hovered))
    result = imgui.combo(*args, **kwargs)
    imgui.pop_style_color()
    return result


def get_current_monitor(wx, wy, ww=None, wh=None):
    # so we always return something sensible
    monitor = glfw.get_primary_monitor()
    bounds  = get_monitor_work_area(monitor)
    bestoverlap = 0
    for mon in glfw.get_monitors():
        m_bounds = get_monitor_work_area(mon)

        if ww is None or wh is None:
            # check if monitor contains (wx,wy)
            if m_bounds.contains([wx, wy]):
                return mon, m_bounds
        else:
            # check overlap of window with monitor
            mx = m_bounds.position[0]
            my = m_bounds.position[0]
            mw = m_bounds.size[0]
            mh = m_bounds.size[1]
            overlap = \
                max(0, min(wx + ww, mx + mw) - max(wx, mx)) * \
                max(0, min(wy + wh, my + mh) - max(wy, my))

            if bestoverlap < overlap:
                bestoverlap = overlap
                monitor = mon
                bounds = m_bounds

    return monitor, bounds

def get_monitor_work_area(monitor):
    monitor_area = glfw.get_monitor_workarea(monitor)
    return hello_imgui.ScreenBounds(monitor_area[:2], monitor_area[2:])

def adjust_bounds_for_framesize(w_bounds, frame_size):
    if platform.os==platform.Os.Windows:
        pos = [
            w_bounds.position[0],
            w_bounds.position[1] + frame_size[1]
        ]
        size = [
            w_bounds.size[0],
            w_bounds.size[1] - frame_size[1]
        ]
    else:
        pos = [
            w_bounds.position[0] + frame_size[0],
            w_bounds.position[1] + frame_size[1],
        ]
        size = [
            w_bounds.size[0] - (frame_size[0]+frame_size[2]),
            w_bounds.size[1] - (frame_size[1]+frame_size[3])
        ]

    return hello_imgui.ScreenBounds(pos, size)


from typing import Callable
from enum import Enum, auto
from imgui_bundle import hello_imgui, imgui, icons_fontawesome_6 as ifa6

from . import utils

icon_font = None

class MsgBox(Enum):
    question= auto()
    info    = auto()
    warn    = auto()
    error   = auto()

def msgbox(title: str, msg: str, type: MsgBox = None, buttons: dict[str, Callable] = True, more: str = None):
    def popup_content():
        spacing = 2 * imgui.get_style().item_spacing.x
        if type is MsgBox.question:
            icon = ifa6.ICON_FA_CIRCLE_QUESTION
            color = (0.45, 0.09, 1.00)
        elif type is MsgBox.info:
            icon = ifa6.ICON_FA_CIRCLE_INFO
            color = (0.10, 0.69, 0.95)
        elif type is MsgBox.warn:
            icon = ifa6.ICON_FA_TRIANGLE_EXCLAMATION
            color = (0.95, 0.69, 0.10)
        elif type is MsgBox.error:
            icon = ifa6.ICON_FA_TRIANGLE_EXCLAMATION
            color = (0.95, 0.22, 0.22)
        else:
            icon = None
        if icon:
            imgui.push_font(icon_font)
            icon_size = imgui.calc_text_size(icon)
            imgui.text_colored((*color,1.),icon)
            imgui.pop_font()
            imgui.same_line(spacing=spacing)
        imgui.begin_group()
        msg_size_y = imgui.calc_text_size(msg).y
        if more:
            msg_size_y += imgui.get_text_line_height_with_spacing() + imgui.get_frame_height_with_spacing()
        if icon and (diff := icon_size.y - msg_size_y) > 0:
            imgui.dummy((0, diff / 2 - imgui.get_style().item_spacing.y))
        imgui.text_unformatted(msg)
        if more:
            imgui.text("")
            if imgui.tree_node_ex("More info", flags=imgui.TreeNodeFlags_.span_avail_width):
                size = imgui.get_io().display_size
                more_size = imgui.calc_text_size(more)
                _26 = hello_imgui.dpi_window_size_factor()*26 + imgui.get_style().scrollbar_size
                width = min(more_size.x + _26, size.x * 0.8 - icon_size.x)
                height = min(more_size.y + _26, size.y * 0.7 - msg_size_y)
                imgui.input_text_multiline(f"###more_info_{title}", more, (width, height), flags=imgui.InputTextFlags_.read_only)
                imgui.tree_pop()
        imgui.end_group()
        imgui.same_line(spacing=spacing)
        imgui.dummy((0, 0))
    return utils.popup(title, popup_content, buttons, closable=False, escape=True, outside=False)



from enum import Enum, auto

from . import utils

class Type(Enum):
    Point       = auto()
    Interval    = auto()

class Event(utils.AutoName):
    Validate    = auto()    # interval to be used for running glassesValidator
    Sync_Camera = auto()    # point to be used for synchronizing different cameras
    Sync_ET_Data= auto()    # episode to be used for synchronization of eye tracker data to scene camera (e.g. using VOR)
    Trial       = auto()    # episode for which to map gaze to plane(s): output for files to be provided to user
events = [x.value for x in Event]
utils.register_type(utils.CustomTypeEntry(Event,'__enum.Event__',str, lambda x: getattr(Event, x.split('.')[1])))

type_map = {
    Event.Validate    : Type.Interval,
    Event.Sync_Camera : Type.Point,
    Event.Sync_ET_Data: Type.Interval,
    Event.Trial       : Type.Interval,
}

tooltip_map = {
    Event.Validate    : 'Validation episode',
    Event.Sync_Camera : 'Camera sync point',
    Event.Sync_ET_Data: 'Eye tracker synchronization episode',
    Event.Trial       : 'Trial',
}



import numpy as np
import cv2

from . import marker

def to_norm_pos(x,y,bbox):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a normalized position
    # in an image of the plane, given the image's bounding box in
    # world units
    # for input (0,0) is bottom left, for output (0,0) is top left
    # bbox is [left, top, right, bottom]

    extents = [bbox[2]-bbox[0], bbox[1]-bbox[3]]
    pos     = [(x-bbox[0])/extents[0], (bbox[1]-y)/extents[1]]    # bbox[1]-y instead of y-bbox[3] to flip y
    return pos

def to_image_pos(x,y,bbox,img_size,margin=[0,0]):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a pixel position in the
    # image, given the image's bounding box in world units
    # imSize should be active image area in pixels, excluding margin

    # fractional position between bounding box edges, (0,0) in bottom left
    pos = to_norm_pos(x,y, bbox)
    # turn into int, add margin
    pos = [p*s+m for p,s,m in zip(pos,img_size,margin)]
    return pos

def in_bbox(x,y,bbox,margin=0.):
    pos = to_norm_pos(x,y,bbox)
    return (pos[0]>=-margin and pos[0]<=1+margin) and (pos[1]>=-margin and pos[1]<=1+margin)

def dist_from_bbox(x,y,bbox):
    pos = to_norm_pos(x,y,bbox)
    if (pos[0]>=0 and pos[0]<=1) and (pos[1]>=0 and pos[1]<=1):
        return 0.   # inside bbox
    # compute max distance from edge of bbox
    dx = pos[0] if pos[0]<0. else pos[0]-1
    dy = pos[1] if pos[1]<0. else pos[1]-1
    return abs(max(dx,dy))


def estimate_homography_known_marker(known: list[marker.Marker], detected_corners, detected_IDs):
    # collect matching corners in image and in world
    img_points = []
    obj_points = []
    detected_IDs = detected_IDs.flatten()
    if len(detected_IDs) != len(detected_corners):
        raise ValueError('unexpected number of IDs (%d) given number of corner arrays (%d)' % (len(detected_IDs),len(detected_corners)))
    for i in range(0, len(detected_IDs)):
        if detected_IDs[i] in known:
            dc = detected_corners[i]
            if dc.shape[0]==1 and dc.shape[1]==4:
                dc = np.reshape(dc,(4,1,2))
            img_points.extend( [x.flatten() for x in dc] )
            obj_points.extend( known[detected_IDs[i]].corners )

    if len(img_points) < 4:
        return None, False

    # compute Homography
    return estimate_homography(obj_points, img_points)

def estimate_homography(obj_points, img_points):
    img_points = np.float32(img_points)
    obj_points = np.float32(obj_points)
    h, _ = cv2.findHomography(img_points, obj_points)
    return h, True

def apply_homography(H, x, y):
    if np.isnan(x) or np.isnan(y):
        return np.full((2,), np.nan)

    src = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    dst = cv2.perspectiveTransform(src,H)
    return dst.flatten()


def distort_point(x, y, camera_matrix, dist_coeff):
    if np.isnan(x) or np.isnan(y):
        return np.full((2,), np.nan)

    # unproject, ignoring distortion as this is an undistored point
    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, camera_matrix, np.asarray([[0.0, 0.0, 0.0, 0.0, 0.0]]))
    points_3d = cv2.convertPointsToHomogeneous(points_2d)
    points_3d.shape = -1, 3

    # reproject, applying distortion
    points_2d, _ = cv2.projectPoints(points_3d, np.zeros((1, 1, 3)), np.zeros((1, 1, 3)), camera_matrix, dist_coeff)
    return points_2d.flatten()

def undistort_point(x, y, camera_matrix, dist_coeff):
    if np.isnan(x) or np.isnan(y):
        return np.full((2,), np.nan)

    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, camera_matrix, dist_coeff, P=camera_matrix) # P=cameraMatrix to reproject to camera
    return points_2d.flatten()

def unproject_point(x, y, camera_matrix, dist_coeff):
    if np.isnan(x) or np.isnan(y):
        return np.full((3,), np.nan)

    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, camera_matrix, dist_coeff)
    points_3d = cv2.convertPointsToHomogeneous(points_2d)
    points_3d.shape = -1, 3
    return points_3d.flatten()


def intersect_plane_ray(plane_normal, plane_point, ray_direction, ray_point, epsilon=1e-6):
    # from https://rosettacode.org/wiki/Find_the_intersection_of_a_line_with_a_plane#Python

    ndotu = plane_normal.dot(ray_direction)
    if abs(ndotu) < epsilon:
        # raise RuntimeError("no intersection or line is within plane")
        return np.full((3,), np.nan)

    w = ray_point - plane_point
    si = -plane_normal.dot(w) / ndotu
    return w + si * ray_direction + plane_point


def angle_between(v1, v2):
    return (180.0 / np.pi) * np.arctan2(np.linalg.norm(np.cross(v1,v2)), np.dot(v1,v2))


import numpy as np
import cv2
import pathlib
from typing import Any, Callable
from enum import Enum, auto

from . import annotation, drawing, intervals, marker, ocv, plane, timestamps
from .gui import video_player


def deploy_marker_images(output_dir: str|pathlib.Path, size: int, ArUco_dict: int=cv2.aruco.DICT_4X4_250, markerBorderBits: int=1):
    # Load the predefined dictionary
    dictionary = cv2.aruco.getPredefinedDictionary(ArUco_dict)

    # Generate the markers
    for i in range(dictionary.bytesList.shape[0]):
        markerImage = np.zeros((size, size), dtype=np.uint8)
        markerImage = cv2.aruco.generateImageMarker(dictionary, i, size, markerImage, markerBorderBits)

        cv2.imwrite(output_dir / f"{i}.png", markerImage)

class ArUcoDetector():
    def __init__(self, ArUco_dict: cv2.aruco.Dictionary, params: dict[str]):
        self._det_params                          = cv2.aruco.DetectorParameters()
        self._det_params.cornerRefinementMethod   = cv2.aruco.CORNER_REFINE_SUBPIX    # good default, user can override
        for p in params:
            setattr(self._det_params, p, params[p])

        self._dict                      = ArUco_dict
        self._det                       = cv2.aruco.ArucoDetector(self._dict, self._det_params)
        self._board: cv2.aruco.Board    = None

        self._camera_params             = ocv.CameraParams(None, None)

    def create_board(self, board_corner_points, ids):
        self._board = create_board(board_corner_points, ids, self._dict)

    def set_board(self, board):
        self._board = board

    def set_intrinsics(self, camera_params: ocv.CameraParams):
        self._camera_params = camera_params

    def detect_markers(self, image: cv2.UMat, min_nmarker_refine = 3):
        corners, ids, rejected_corners = _detect_markers(image, self._det)
        recovered_ids = None

        # Refine detected markers (eliminates markers not part of our poster, adds missing markers to the poster)
        if self._board and ids is not None and min_nmarker_refine and len(ids)>min_nmarker_refine:
            corners, ids, rejected_corners, recovered_ids = \
                self.refine_detected_markers(image, corners, ids, rejected_corners)
        return corners, ids, rejected_corners, recovered_ids

    def refine_detected_markers(self, image, detected_corners, detected_ids, rejected_corners):
        return _refine_detection(image, detected_corners, detected_ids, rejected_corners, self._det, self._board, self._camera_params.camera_mtx, self._camera_params.distort_coeffs)

    def _match_image_points(self, corners, ids):
        return self._board.matchImagePoints(corners, ids) # -> objP, imgP

    def estimate_pose(self, corners, ids) -> tuple[int, np.ndarray, np.ndarray, float]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_pose_impl(objP, imgP)

    def _estimate_pose_impl(self, objP, imgP):
        # NB: N_markers also flags success of the pose estimation. Also 0 if not successful or not possible (missing intrinsics)
        N_markers, R_vec, T_vec, reprojection_error = 0, None, None, -1.
        if objP is None or not self._camera_params.has_intrinsics():
            return N_markers, R_vec, T_vec, reprojection_error

        n_solutions, R_vec, T_vec, reprojection_error = \
            cv2.solvePnPGeneric(objP, imgP, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, np.empty(1), np.empty(1))
        if n_solutions:
            N_markers = int(objP.shape[0]/4)
        return N_markers, R_vec[0], T_vec[0], reprojection_error[0][0]

    def estimate_homography(self, corners, ids) -> tuple[np.ndarray, bool]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_homography_impl(objP, imgP)

    def _estimate_homography_impl(self, objP, imgP):
        # NB: N_markers also flags success of the pose estimation. Also 0 if not successful or not possible (missing intrinsics)
        from . import transforms
        N_markers, H = 0, None
        if objP is None:
            return N_markers, H

        # use undistorted marker corners if possible
        if self._camera_params.has_intrinsics():
            imgP = np.vstack([cv2.undistortPoints(x, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, P=self._camera_params.camera_mtx) for x in imgP])

        H, status = transforms.estimate_homography(objP, imgP)
        if status:
            N_markers = int(objP.shape[0]/4)
        return N_markers, H

    # higher level functions for detecting + pose estimation, and for results visualization
    def estimate_pose_and_homography(self, frame_idx, min_num_markers, corners, ids) -> tuple[plane.Pose, dict[str]]:
        pose = plane.Pose(frame_idx)
        if ids is not None and len(ids) >= min_num_markers:
            # get matching image and board points
            objP, imgP = self._match_image_points(corners, ids)

            # get camera pose
            pose.pose_N_markers, pose.pose_R_vec, pose.pose_T_vec, pose.pose_reprojection_error = \
                self._estimate_pose_impl(objP, imgP)

            # also get homography (direct image plane to plane in world transform)
            pose.homography_N_markers, pose.homography_mat = \
                self._estimate_homography_impl(objP, imgP)
        return pose

    def detect_and_estimate(self, frame, frame_idx, min_num_markers) -> tuple[plane.Pose, dict[str]]:
        corners, ids, rejectedImgPoints, recoveredIds = self.detect_markers(frame, min_nmarker_refine=min_num_markers)
        pose = self.estimate_pose_and_homography(frame_idx, min_num_markers, corners, ids)
        return pose, {'corners': corners, 'ids': ids, 'rejectedImgPoints': rejectedImgPoints, 'recoveredIds': recoveredIds}

    def visualize(self, frame, pose: plane.Pose, detect_dict, arm_length, sub_pixel_fac = 8, show_detected_markers = True, show_board_axis = True, show_rejected_markers = False, special_highlight = None):
        if special_highlight is None:
            special_highlight = []
        # for debug, can draw rejected markers on frame
        if show_rejected_markers:
            cv2.aruco.drawDetectedMarkers(frame, detect_dict['rejectedImgPoints'], None, borderColor=(0,0,255))

        # if any markers were detected, draw where on the frame
        if show_detected_markers:
            special_highlight += [detect_dict['recoveredIds'],(255,255,0)]
            drawing.arucoDetectedMarkers(frame, detect_dict['corners'], detect_dict['ids'], sub_pixel_fac=sub_pixel_fac, special_highlight=special_highlight)

        if show_board_axis:
            if pose.pose_successful():
                # draw axis indicating plane pose (origin and orientation)
                drawing.openCVFrameAxis(frame, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, pose.pose_R_vec, pose.pose_T_vec, arm_length, 3, sub_pixel_fac)

            if pose.homography_successful():
                # find where plane origin is expected to be in the image
                target = pose.plane_to_cam_homography([0., 0.], self._camera_params)
                # draw target location on image
                if target[0] >= 0 and target[0] < frame.shape[1] and target[1] >= 0 and target[1] < frame.shape[0]:
                    drawing.openCVCircle(frame, target, 3, (0,0,0), -1, sub_pixel_fac)




def _detect_markers(image: cv2.UMat, det: cv2.aruco.ArucoDetector):
    corners, ids, rejected_corners = det.detectMarkers(image)
    if np.any(ids==None):
        ids = None
    return corners, ids, rejected_corners

def _refine_detection(image: cv2.UMat, detected_corners, detected_ids, rejected_corners, det: cv2.aruco.ArucoDetector, board: cv2.aruco.Board, camera_mtx, distort_coeffs):
    corners, ids, rejectedImgPoints, recoveredIds = \
        det.refineDetectedMarkers(
            image = image, board = board,
            detectedCorners = detected_corners, detectedIds = detected_ids, rejectedCorners = rejected_corners,
            cameraMatrix = camera_mtx, distCoeffs = distort_coeffs
            )
    if corners and corners[0].shape[0]==4:
        # there are versions out there where there is a bug in output shape of each set of corners, fix up
        corners = [np.reshape(c,(1,4,2)) for c in corners]
    if rejectedImgPoints and rejectedImgPoints[0].shape[0]==4:
        # same as for corners
        rejectedImgPoints = [np.reshape(c,(1,4,2)) for c in rejectedImgPoints]

    return corners, ids, rejectedImgPoints, recoveredIds

def _refine_for_multiple_planes(image, detected_corners, detected_ids, rejected_corners, detectors: dict[str, ArUcoDetector], camera_mtx, distort_coeffs):
    result = {}
    for d in detectors:
        a,b,c,e = _refine_detection(image, detected_corners, detected_ids, rejected_corners, detectors[d]._det, detectors[d]._board, detectors[d]._camera_params.camera_mtx, detectors[d]._camera_params.distort_coeffs)
        result[d] = dict(zip(['corners', 'ids', 'rejectedImgPoints', 'recoveredIds'],(a,b,c,e)))
    return result

def create_board(board_corner_points: list[np.ndarray], ids: list[int], ArUco_dict: cv2.aruco.Dictionary):
    board_corner_points = np.dstack(board_corner_points)        # list of 2D arrays -> 3D array
    board_corner_points = np.rollaxis(board_corner_points,-1)   # 4x2xN -> Nx4x2
    board_corner_points = np.pad(board_corner_points,((0,0),(0,0),(0,1)),'constant', constant_values=(0.,0.)) # Nx4x2 -> Nx4x3
    return cv2.aruco.Board(board_corner_points, ArUco_dict, np.array(ids))

class Status(Enum):
    Ok = auto()
    Skip = auto()
    Finished = auto()

class PoseEstimator:
    def __init__(self, video_file: str|pathlib.Path, frame_timestamp_file: str|pathlib.Path|timestamps.VideoTimestamps, camera_calibration_file: str|pathlib.Path|ocv.CameraParams):
        self.video_ts   = frame_timestamp_file if isinstance(frame_timestamp_file,timestamps.VideoTimestamps) else timestamps.VideoTimestamps(frame_timestamp_file)
        self.video      = ocv.CV2VideoReader(video_file, self.video_ts.timestamps)
        self.cam_params = camera_calibration_file if isinstance(camera_calibration_file,ocv.CameraParams) else ocv.CameraParams.read_from_file(camera_calibration_file)

        self.planes                 : list[str]                             = []
        self.plane_setups           : dict[str, dict[str]]                  = {}
        self.plane_proc_intervals   : dict[str, list[int]|list[list[int]]]  = {}
        self._aruco_boards          : dict[str, cv2.aruco.Board]            = {}
        self._all_aruco_ids         : set[int]                              = set()
        self._detectors             : dict[str, ArUcoDetector]              = {}
        self._single_detect_pass                                            = False
        self.allow_early_exit                                               = True

        self.individual_markers                 : dict[int, dict[str]]  = {}
        self._individual_marker_object_points   : dict[int, np.ndarray] = {}
        self.proc_individial_markers_all_frames                         = False

        self.extra_proc_functions   : dict[str, Callable[[np.ndarray,Any], Any]]    = {}
        self.extra_proc_intervals   : dict[str, list[int]|list[list[int]]]          = {}
        self.extra_proc_parameters  : dict[str]                                     = {}
        self.extra_proc_visualizer  : dict[str, Callable[[np.ndarray,Any], None]]   = {}

        self.gui                    : video_player.GUI  = None
        self.has_gui                                    = False

        self.do_visualize                               = False
        self.sub_pixel_fac                              = 8
        self.show_detected_markers                      = True
        self.show_board_axes                            = True
        self.show_individual_marker_axes                = True
        self.show_sync_func_output                      = True
        self.show_unexpected_markers                    = True
        self.show_rejected_markers                      = False

        self._first_frame       = True
        self._do_report_frames  = True

    def __del__(self):
        if self.has_gui:
            self.gui.stop()

    def add_plane(self, plane: str, planes_setup: dict[str], processing_intervals: list[int]|list[list[int]] = None):
        if plane in self.planes:
            raise ValueError(f'Cannot register the plane "{plane}", it is already registered')
        self.planes.append(plane)
        self.plane_setups[plane] = planes_setup
        self.plane_proc_intervals[plane] = processing_intervals

        self._aruco_boards[plane]   = planes_setup['plane'].get_aruco_board()
        self._all_aruco_ids.update(self._aruco_boards[plane].getIds())
        self._detectors[plane]      = ArUcoDetector(self._aruco_boards[plane].getDictionary(), planes_setup['aruco_params'])
        self._detectors[plane].set_board(self._aruco_boards[plane])
        self._detectors[plane].set_intrinsics(self.cam_params)

        self._cache: tuple[Status, dict[str, plane.Pose], dict[str, marker.Pose], dict[str, list[int, Any]], tuple[np.ndarray, int, float]] = None  # self._cache[4][1] is frame number

        # check if we can do an optimization of detecting the markers only once for multiple planes (if it makes sense because we have more than one plane)
        aruco_dicts = [self.plane_setups[p]['aruco_dict'] for p in self.planes if 'aruco_dict' in self.plane_setups[p]]
        self._single_detect_pass = len(self.planes)>1 and len(aruco_dicts)==len(self.planes) and len(set(aruco_dicts))==1 and all([self.plane_setups[self.planes[0]]['aruco_params']==self.plane_setups[p]['aruco_params'] for p in self.planes])

    def add_individual_marker(self, marker_id: int, marker_setup):
        if not self._single_detect_pass and len(self.planes)!=1:
            raise ValueError("Detecting and reporting individual markers is only supported when there is a single plane, or all planes have identical ArUco setup")
        if marker_id in self.individual_markers:
            raise ValueError(f'Cannot register the individual marker with id {marker_id}, it is already registered')
        self.individual_markers[marker_id] = marker_setup
        marker_size = self.individual_markers[marker_id]['marker_size']
        self._individual_marker_object_points[marker_id] = np.array([[-marker_size/2,  marker_size/2, 0],
                                                                     [ marker_size/2,  marker_size/2, 0],
                                                                     [ marker_size/2, -marker_size/2, 0],
                                                                     [-marker_size/2, -marker_size/2, 0]])
        self._all_aruco_ids.add(marker_id)

    def register_extra_processing_fun(self,
                                      name: str,
                                      func: Callable[[np.ndarray,Any], Any],
                                      processing_intervals: list[int]|list[list[int]],
                                      func_parameters: dict[str],
                                      visualizer: Callable[[np.ndarray, Any], None]):
        if name in self.extra_proc_functions:
            raise ValueError(f'Cannot register the extra processing function "{name}", it is already registered')
        self.extra_proc_functions[name] = func
        self.extra_proc_intervals[name] = processing_intervals
        self.extra_proc_parameters[name]= func_parameters
        self.extra_proc_visualizer[name]= visualizer

    def attach_gui(self, gui: video_player.GUI, episodes: dict[annotation.Event, list[int]] = None, window_id: int = None):
        self.gui                    = gui
        self.has_gui                = self.gui is not None
        self.do_visualize           = self.has_gui

        if self.has_gui:
            self.gui.set_show_timeline(True, self.video_ts, episodes, window_id)

    def set_allow_early_exit(self, allow_early_exit: bool):
        # if False, processing will not stop because last frame with a defined plane or extra processing is reached
        self.allow_early_exit = allow_early_exit

    def set_visualize_on_frame(self, do_visualize: bool):
        self.do_visualize           = do_visualize

    def set_do_report_frames(self, do_report_frames: bool):
        self._do_report_frames = do_report_frames

    def get_video_info(self) -> tuple[int, int, float]:
        return int(self.video.get_prop(cv2.CAP_PROP_FRAME_WIDTH)), \
               int(self.video.get_prop(cv2.CAP_PROP_FRAME_HEIGHT)), \
                   self.video.get_prop(cv2.CAP_PROP_FPS)

    def process_one_frame(self, wanted_frame_idx:int = None) -> tuple[Status, dict[str, plane.Pose], dict[str, marker.Pose], dict[str, list[int, Any]], tuple[np.ndarray, int, float]]:
        if self._first_frame and self.has_gui:
            self.gui.set_playing(True)

        if wanted_frame_idx is not None and self._cache is not None and self._cache[4][1]==wanted_frame_idx:
            return self._cache

        should_exit, frame, frame_idx, frame_ts = self.video.read_frame(report_gap=True, wanted_frame_idx=wanted_frame_idx)

        if should_exit or (self.allow_early_exit and not self.proc_individial_markers_all_frames and \
            (
                intervals.beyond_last_interval(frame_idx, self.plane_proc_intervals) and \
                (not self.extra_proc_intervals or intervals.beyond_last_interval(frame_idx, self.extra_proc_intervals))
            )):
            self._cache = Status.Finished, None, None, None, (None, None, None)
            return self._cache
        if self._do_report_frames:
            self.video.report_frame()

        if self.has_gui:
            if self._first_frame and frame is not None:
                self.gui.set_frame_size(frame.shape)
                self._first_frame = False

            requests = self.gui.get_requests()
            for r,_ in requests:
                if r=='exit':   # only request we need to handle
                    should_exit = True
                    self._cache = Status.Finished, None, None, None, (None, None, None)
                    return self._cache

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.set(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        planes_for_this_frame = [p for p in self.planes if intervals.is_in_interval(frame_idx, self.plane_proc_intervals[p])]
        extra_processing_for_this_frame = [e for e in self.extra_proc_functions if intervals.is_in_interval(frame_idx, self.extra_proc_intervals[e])]
        if frame is None or (not (self.proc_individial_markers_all_frames and self.individual_markers) and not planes_for_this_frame and not extra_processing_for_this_frame):
            # we don't have a valid frame or nothing to do, continue to next
            if self.has_gui:
                # do update timeline of the viewers
                self.gui.update_image(None, frame_ts/1000., frame_idx)
            self._cache = Status.Skip, None, None, None, (frame, frame_idx, frame_ts)
            return self._cache

        pose_out                : dict[str, plane.Pose]     = {}
        individual_marker_out   : dict[str, marker.Pose]    = {}
        extra_processing_out    : dict[str, list[int, Any]] = {}
        if planes_for_this_frame or (self.proc_individial_markers_all_frames and self.individual_markers):
            # detect markers
            detect_dicts = {}
            if self._single_detect_pass or self.individual_markers:
                corners, ids, rejected_corners = _detect_markers(frame, self._detectors[self.planes[0]]._det)
                if planes_for_this_frame:
                    detect_dicts = _refine_for_multiple_planes(frame, corners, ids, rejected_corners, {p:self._detectors[p] for p in planes_for_this_frame}, self.cam_params.camera_mtx, self.cam_params.distort_coeffs)
            else:
                for p in planes_for_this_frame:
                    detect_dicts[p] = dict(zip(['corners', 'ids', 'rejectedImgPoints', 'recoveredIds'], self._detectors[p].detect_markers(frame, self.plane_setups[p]['min_num_markers'])))
            # determine pose
            for p in planes_for_this_frame:
                pose = self._detectors[p].estimate_pose_and_homography(frame_idx, self.plane_setups[p]['min_num_markers'], detect_dicts[p]['corners'], detect_dicts[p]['ids'])
                pose_out[p] = pose

            # deal with individual markers, if any
            if self.individual_markers and ids is not None:
                found_markers = np.where([x[0] in self.individual_markers for x in ids])[0]
                for idx in found_markers:
                    m_id = ids[idx][0]
                    pose = marker.Pose(frame_idx)
                    if self.cam_params.has_intrinsics():
                        # can only get marker pose if we have a calibrated camera (need intrinsics), else at least flag that marker was found
                        _, pose.R_vec, pose.T_vec = cv2.solvePnP(self._individual_marker_object_points[m_id], corners[idx], self.cam_params.camera_mtx, self.cam_params.distort_coeffs, flags=cv2.SOLVEPNP_IPPE_SQUARE)
                    individual_marker_out[m_id] = pose

        for e in extra_processing_for_this_frame:
            extra_processing_out[e] = [frame_idx, *self.extra_proc_functions[e](frame,**self.extra_proc_parameters[e])]

        # now that all processing is done, handle visualization, if any
        if self.do_visualize:
            # for visualization purposes, either filter detected markers so only the expected ones
            # are shown (self.show_unknown_markers==False), or show them in a different color (self.show_unknown_markers==True)
            to_highlight = set()
            # get expected markers
            expected = set()
            if planes_for_this_frame:
                for p in planes_for_this_frame:
                    expected.update(self._detectors[p]._board.getIds())
                for m_id in self.individual_markers:
                    expected.add(m_id)
            else:
                expected = self._all_aruco_ids
            # filter dicts to find unexpected markers
            for p in detect_dicts:
                if detect_dicts[p]['ids'] is None:
                    continue
                unexpected = set(detect_dicts[p]['ids'].flatten())-expected
                if not unexpected:
                    continue
                if self.show_unexpected_markers:
                    to_highlight |= unexpected
                else:
                    to_remove = np.where([x[0] in unexpected for x in detect_dicts[p]['ids']])[0]
                    detect_dicts[p]['ids'] = np.delete(detect_dicts[p]['ids'], to_remove, axis=0)
                    detect_dicts[p]['corners'] = tuple(v for i,v in enumerate(detect_dicts[p]['corners']) if i not in to_remove)
            special_highlights = None
            if to_highlight:
                special_highlights = [list(to_highlight), (150,253,253)]
            # now actually draw them
            for p in planes_for_this_frame:
                # draw detection and pose, if wanted
                self._detectors[p].visualize(frame, pose_out[p], detect_dicts[p], self.plane_setups[p]['plane'].marker_size/2, self.sub_pixel_fac, self.show_detected_markers, self.show_board_axes, self.show_rejected_markers, special_highlights)

            # ensure visualization of detected and rejected markers is honored when marker detection
            # was done but self._detectors[p].visualize() above won't be called because there are no
            # planes for this frame
            if not planes_for_this_frame and (self.proc_individial_markers_all_frames or self.individual_markers):
                if self.show_rejected_markers:
                    cv2.aruco.drawDetectedMarkers(frame, rejected_corners, None, special=(0,0,255))
                if self.show_detected_markers:
                    drawing.arucoDetectedMarkers(frame, corners, ids, sub_pixel_fac=self.sub_pixel_fac)

            # deal with individual markers, if any
            for m_id in individual_marker_out:
                if self.show_detected_markers:
                    # draw the detected marker in a different color
                    idx = np.where(ids==m_id)[0][0]
                    drawing.arucoDetectedMarkers(frame, [corners[idx]], ids[idx].reshape((1,1)), sub_pixel_fac=self.sub_pixel_fac, special_highlight=[[m_id],(255,0,255)])
                if self.show_individual_marker_axes and self.cam_params.has_intrinsics():
                    individual_marker_out[m_id].draw_frame_axis(frame, self.cam_params, self.individual_markers[m_id]['marker_size']/2, self.sub_pixel_fac)

            # draw output of extra processing functions
            for e in extra_processing_for_this_frame:
                if self.show_sync_func_output and self.extra_proc_visualizer[e]:
                    self.extra_proc_visualizer[e](frame, *extra_processing_out[e])

        if self.has_gui:
            self.gui.update_image(frame, frame_ts/1000., frame_idx)

        self._cache = Status.Ok, pose_out, individual_marker_out, extra_processing_out, (frame, frame_idx, frame_ts)
        return self._cache

    def process_video(self) -> tuple[dict[str, list[plane.Pose]], dict[str, list[marker.Pose]], dict[str, list[list[int, Any]]]]:
        poses_out               : dict[str, list[plane.Pose]]       = {p:[] for p in self.planes}
        individual_markers_out  : dict[str, list[marker.Pose]]      = {i:[] for i in self.individual_markers}
        extra_processing_out    : dict[str, list[list[int, Any]]]   = {e:[] for e in self.extra_proc_functions}
        while True:
            status, pose, individual_marker, extra_proc, _ = self.process_one_frame()
            if status==Status.Finished:
                break
            if status==Status.Skip:
                continue
            # store outputs
            for p in pose:
                poses_out[p].append(pose[p])
            for i in individual_marker:
                individual_markers_out[i].append(individual_marker[i])
            for e in extra_proc:
                extra_processing_out[e].append(extra_proc[e])

        return poses_out, individual_markers_out, extra_processing_out



gaze_data_fname                 = 'gazeData.tsv'
frame_timestamps_fname          = 'frameTimestamps.tsv'
scene_camera_video_fname_stem   = 'worldCamera'
scene_camera_calibration_fname  = 'calibration.xml'


import numpy as np
import pandas as pd
import cv2
import pathlib
import bisect
import warnings


class CameraParams:
    def __init__(self,
                 # camera info
                 resolution: np.ndarray,
                 # intrinsics
                 camera_mtx: np.ndarray, distort_coeffs: np.ndarray = None,
                 # extrinsics
                 rotation_vec: np.ndarray = None, position: np.ndarray = None):

        self.resolution     : np.ndarray = resolution
        self.camera_mtx     : np.ndarray = camera_mtx
        self.distort_coeffs : np.ndarray = distort_coeffs
        self.rotation_vec   : np.ndarray = rotation_vec
        self.position       : np.ndarray = position

    @staticmethod
    def read_from_file(fileName: str|pathlib.Path) -> 'CameraParams':
        fileName = pathlib.Path(fileName)
        if not fileName.is_file():
            return CameraParams(None,None)

        fs = cv2.FileStorage(fileName, cv2.FILE_STORAGE_READ)
        resolution      = fs.getNode("resolution").mat()
        cameraMatrix    = fs.getNode("cameraMatrix").mat()
        distCoeff       = fs.getNode("distCoeff").mat()
        # camera extrinsics for 3D gaze
        cameraRotation  = fs.getNode("rotation").mat()
        if cameraRotation is not None:
            cameraRotation  = cv2.Rodrigues(cameraRotation)[0]  # need rotation vector, not rotation matrix
        cameraPosition  = fs.getNode("position").mat()
        fs.release()

        return CameraParams(resolution,cameraMatrix,distCoeff,cameraRotation,cameraPosition)

    def has_intrinsics(self):
        return (self.camera_mtx is not None) and (self.distort_coeffs is not None)


class CV2VideoReader:
    def __init__(self, file: str|pathlib.Path, timestamps: list|np.ndarray|pd.DataFrame):
        self.file = pathlib.Path(file)
        if isinstance(timestamps,list):
            self.ts = np.array(timestamps)
        elif isinstance(timestamps, pd.DataFrame):
            self.ts = timestamps['timestamp'].to_numpy()
        else:
            self.ts = timestamps

        self.cap = cv2.VideoCapture(self.file)
        if not self.cap.isOpened():
            raise RuntimeError('the file "{}" could not be opened'.format(str(self.file)))
        self.nframes = len(self.ts)
        self.frame_idx = -1
        self._last_good_ts = (-1, -1., -1.)  # frame_idx, ts from opencv, ts from file
        self._cache: tuple[bool, np.ndarray, int, float] = None # self._cache[2] is frame index

    def __del__(self):
        self.cap.release()

    def get_prop(self, cv2_prop):
        return self.cap.get(cv2_prop)

    def set_prop(self, cv2_prop, val):
        return self.cap.set(cv2_prop, val)

    # NB: we seek by spooling, because I found seeking through setting cv2.CAP_PROP_POS_MSEC unreliable
    def read_frame(self, report_gap=False, wanted_frame_idx:int=None) -> tuple[bool, np.ndarray, int, float]:
        if wanted_frame_idx!=None:
            if wanted_frame_idx<0 or wanted_frame_idx>=self.nframes:
                raise ValueError(f'wanted_frame_idx ({wanted_frame_idx}) out of bounds ([0-{self.nframes-1}])')
        else:
            wanted_frame_idx = self.frame_idx+1

        if self.frame_idx>wanted_frame_idx:
            warnings.warn(f'Requested frame ({wanted_frame_idx}) was earlier than current position of reader (frame {self.frame_idx}). Impossible to deliver because this video reader strictly advances forward. Returning last read frame', RuntimeWarning)
            # this condition can only occur if we've already read something and thus have a cache, so this check should never trigger
            if self._cache is None:
                raise RuntimeError(f'No cache, unexpected failure mode, contact developer')
            return self._cache
        elif self._cache is not None and self._cache[2]==wanted_frame_idx:
            return self._cache

        while True:
            ret, frame = self.cap.read()
            # get timestamp of this frame according to OpenCV
            # NB: this timestamp does not take into account edit lists in an mp4
            # it seems (compare with output of .\ffprobe.exe file.mp4 -select_streams 0 -show_entries frame=pts_time)
            # which is one of the reasons we provide our own timestamps
            ocv_ts = self.cap.get(cv2.CAP_PROP_POS_MSEC)
            self.frame_idx += 1

            # check if we're done. Can't trust ret==False to indicate we're at end of video, as
            # it may also return False for some corrupted frames that we can just read past
            if not ret and (self.frame_idx==0 or self.frame_idx/self.nframes>.99):
                self._cache = True, None, None, None
                return self._cache

            # keep going
            ts_from_list = self.ts[self.frame_idx]
            if self.frame_idx==1 or ocv_ts>0.:
                # check for gap, and if there is a gap, fix up frame_idx if needed
                if self._last_good_ts[0]!=-1 and ts_from_list-self._last_good_ts[2] < ocv_ts-self._last_good_ts[1]-1:  # little bit of leeway (1ms) for precision or mismatched timestamps
                    # we skipped some frames altogether, need to correct current frame_idx
                    t_jump = ocv_ts-self._last_good_ts[1]   # compare OpenCV timestamps to get size of jump
                    tss = self.ts-self._last_good_ts[2]     # apply jump to our own timestamps (so, we're robust to e.g. OpenCV ignoring the edit list)
                    # find best matching frame idx so we catch up with the jump
                    self.frame_idx = self._find_closest_idx(t_jump, tss)
                    ts_from_list = self.ts[self.frame_idx]
                    if report_gap and self.frame_idx-self._last_good_ts[0]>1:
                        print(f'Frame discontinuity detected (jumped from {self._last_good_ts[0]} to {self.frame_idx}), there are probably corrupt frames in your video')
                self._last_good_ts = (self.frame_idx, ocv_ts, ts_from_list)

            # keep spooling until we arrive at the wanted frame
            if self.frame_idx==wanted_frame_idx:
                if not ret or frame is None:
                    # we might not have a valid frame, but we're not done yet
                    self._cache = False, None,  self.frame_idx, ts_from_list
                else:
                    self._cache = False, frame, self.frame_idx, ts_from_list
                return self._cache

    def _find_closest_idx(self, time: float, times: np.ndarray) -> int:
        idx = bisect.bisect(times, time)
        if abs(times[idx-1]-time)<abs(times[idx]-time):
            idx -= 1
        return idx

    def report_frame(self, interval=100):
        if self.frame_idx%interval==0:
            print('  frame {}'.format(self.frame_idx))


import concurrent
import pathlib
from typing import Callable

from . import async_thread, file_actions, platform

class FileActionProvider:
    def __init__(self, listing_callback: Callable[[str, str|pathlib.Path, list[file_actions.DirEntry]|Exception], None]=None, action_callback: Callable[[str, pathlib.Path, str, pathlib.Path|Exception], None] = None):
        self.waiters: set[concurrent.futures.Future] = set()

        self.listing_callbacks: list[Callable[[list[file_actions.DirEntry]|Exception, bool], None]] = []
        if listing_callback:
            self.listing_callbacks.append(listing_callback)
        self.action_callbacks:  list[Callable[[pathlib.Path|Exception, str], None]] = []
        if action_callback:
            self.action_callbacks.append(action_callback)

    def __del__(self):
        for w in self.waiters:
            if not w.done():
                w.cancel()

    local_name = 'This PC' if platform.os==platform.Os.Windows else 'Root'

    def get_listing(self, path: str|pathlib.Path) -> list[file_actions.DirEntry]|concurrent.futures.Future:
        fut = None
        if platform.os==platform.Os.Windows:
            if path=='root':
                try:
                    result = file_actions.get_drives()
                    result.extend(file_actions.get_thispc_listing())
                except Exception as exc:
                    result = exc
                self._listing_done(result, 'root')
            else:
                # check whether this is a path to a network computer (e.g. \\SERVER)
                net_comp = file_actions.get_net_computer(path)
                try:
                    if net_comp:
                        # network computer name, get its shares
                        result = file_actions.get_visible_shares(net_comp,'Guest','')
                    else:
                        # normal directory or share on a network computer, no special handling needed
                        result = file_actions.get_dir_list_sync(path)
                except Exception as exc:
                    result = exc
                self._listing_done(result, path)
        else:
            try:
                l_path = path
                if path=='root':
                    l_path = '/'
                result = file_actions.get_dir_list_sync(l_path)
            except Exception as exc:
                result = exc
            self._listing_done(result, path)
        if fut:
            self.waiters.add(fut)
        return fut

    def _listing_done(self, fut: concurrent.futures.Future|list[file_actions.DirEntry], path: str|pathlib.Path):
        result = self._get_result_from_future(fut)
        if result=='cancelled':
            return  # nothing more to do

        if not self.listing_callbacks:
            return
        if result is None:
            return

        # call callback
        for c in self.listing_callbacks:
            c(path, result)

    def make_dir(self, path: pathlib.Path):
        action = 'make_dir'
        fut = async_thread.run(file_actions.make_dir(path), lambda f: self._action_done(f, path, action))
        self.waiters.add(fut)
        return fut

    def rename_path(self, old_path: pathlib.Path, new_path: pathlib.Path):
        action = 'rename_path'
        fut = async_thread.run(file_actions.rename_path(old_path, new_path), lambda f: self._action_done(f, old_path, action))
        self.waiters.add(fut)
        return fut

    def _action_done(self, fut: concurrent.futures.Future, path: pathlib.Path, action: str):
        result = self._get_result_from_future(fut)
        if result=='cancelled':
            return  # nothing more to do

        if not self.action_callbacks:
            return

        # call callback
        for c in self.action_callbacks:
            c(path, action, result)

    def _get_result_from_future(self, fut: concurrent.futures.Future|list[file_actions.DirEntry]) -> list[file_actions.DirEntry]:
        if isinstance(fut, concurrent.futures.Future):
            self.waiters.discard(fut)
            try:
                return fut.result()
            except concurrent.futures.CancelledError:
                return 'cancelled'
            except Exception as exc:
                return exc
        else:
            return fut


import pathlib

from .. import annotation, drawing, gaze_headref, gaze_worldref, intervals, ocv, plane, timestamps
from . import video_player


def show_visualization(
        in_video: str|pathlib.Path, frame_timestamp_file: str|pathlib.Path, camera_calibration_file: str|pathlib.Path,
        planes: dict[str, plane.Plane], poses: dict[str, dict[int, plane.Pose]],
        head_gazes: dict[int, list[gaze_headref.Gaze]], plane_gazes: dict[str, dict[int, list[gaze_worldref.Gaze]]],
        annotations: dict[annotation.Event, list[list[int]]],
        gui: video_player.GUI, show_planes: bool, show_only_intervals: bool, sub_pixel_fac: int
    ):
    in_video                = pathlib.Path(in_video)
    frame_timestamp_file    = pathlib.Path(frame_timestamp_file)
    camera_calibration_file = pathlib.Path(camera_calibration_file)

    # prep visualizations
    # open video
    video_ts        = timestamps.VideoTimestamps(frame_timestamp_file)
    cap             = ocv.CV2VideoReader(in_video, video_ts.timestamps)
    cam_params      = ocv.CameraParams.read_from_file(camera_calibration_file)

    # flatten if needed
    annotations_flat: dict[annotation.Event, list[int]] = {}
    for e in annotations:
        if annotations[e] and isinstance(annotations[e][0],list):
            annotations_flat[e] = [i for iv in annotations[e] for i in iv]
        else:
            annotations_flat[e] = annotations[e].copy()
    gui.set_show_timeline(True, video_ts, annotations_flat, window_id=gui.main_window_id)

    # add windows for planes, if wanted
    if show_planes:
        plane_win_id = {p: gui.add_window(p) for p in planes}

    max_frame_idx = max(head_gazes.keys())
    should_exit = False
    first_frame = True
    for frame_idx in range(max_frame_idx+1):
        done, frame, frame_idx, frame_ts = cap.read_frame(report_gap=True)
        if first_frame and frame is not None:
            gui.set_frame_size(frame.shape, gui.main_window_id)
            first_frame = False
        if done or intervals.beyond_last_interval(frame_idx, annotations):
            break

        requests = gui.get_requests()
        for r,p in requests:
            if r=='exit':   # only request we need to handle
                should_exit = True
                break
        if should_exit:
            break

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.get(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        if show_only_intervals and not intervals.is_in_interval(frame_idx, annotations) or frame is None:
            # we don't have a valid frame or no need to show this frame
            # do update timeline of the viewers
            gui.update_image(None, frame_ts/1000., frame_idx, window_id=gui.main_window_id)
            if show_planes:
                for p in planes:
                    gui.update_image(None, frame_ts/1000., frame_idx, window_id=plane_win_id[p])
            continue

        if show_planes:
            ref_img = {p: planes[p].get_ref_image(400) for p in planes}

        if frame_idx in head_gazes:
            for gaze_head in head_gazes[frame_idx]:
                # draw gaze point on scene video
                gaze_head.draw(frame, cam_params, sub_pixel_fac)

        # draw plane gazes on video and plane
        for p in planes:
            if frame_idx in plane_gazes[p]:
                for gaze_world in plane_gazes[p][frame_idx]:
                    gaze_world.draw_on_world_video(frame, cam_params, sub_pixel_fac)
                    if show_planes:
                        gaze_world.draw_on_plane(ref_img[p], planes[p], sub_pixel_fac)

        if show_planes:
            for p in planes:
                gui.update_image(ref_img[p], frame_ts/1000., frame_idx, window_id=plane_win_id[p])

        # if we have plane pose, draw plane origin on video
        for p in planes:
            if frame_idx in poses[p]:
                a = poses[p][frame_idx].get_origin_on_image(cam_params)
                drawing.openCVCircle(frame, a, 3, (0,255,0), -1, sub_pixel_fac)
                ll = 20
                drawing.openCVLine(frame, (a[0],a[1]-ll), (a[0],a[1]+ll), (0,255,0), 1, sub_pixel_fac)
                drawing.openCVLine(frame, (a[0]-ll,a[1]), (a[0]+ll,a[1]), (0,255,0), 1, sub_pixel_fac)

        gui.update_image(frame, frame_ts/1000., frame_idx, window_id=gui.main_window_id)

    gui.stop()


import numpy as np
import pathlib

from . import data_files, drawing, ocv


class Marker:
    def __init__(self, key: int, center: np.ndarray, corners: list[np.ndarray]=None, color: str=None, rot: float=0.):
        self.key = key
        self.center = center
        self.corners = corners
        self.color = color
        self.rot = rot

    def __str__(self):
        ret = '[%d]: center @ (%.2f, %.2f), rot %.0f deg' % (self.key, self.center[0], self.center[1], self.rot)
        return ret

    def shift(self, offset=np.ndarray):
        self.center += offset
        if self.corners:
            for c in self.corners:
                c += offset

def corners_intersection(corners):
    line1 = ( corners[0], corners[2] )
    line2 = ( corners[1], corners[3] )
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return np.array( [x,y] ).astype('float32')

class Pose:
    # description of tsv file used for storage
    _columns_compressed = {'frame_idx': 1, 'R_vec': 3, 'T_vec': 3}
    _non_float          = {'frame_idx': int}

    def __init__(self,
                 frame_idx  : int,
                 R_vec      : np.ndarray= None,
                 T_vec      : np.ndarray= None):
        self.frame_idx  : int         = frame_idx
        self.R_vec      : np.ndarray  = R_vec
        self.T_vec      : np.ndarray  = T_vec

    def draw_frame_axis(self, frame, camera_params: ocv.CameraParams, arm_length, sub_pixel_fac = 8):
        if not camera_params.has_intrinsics():
            return
        drawing.openCVFrameAxis(frame, camera_params.camera_mtx, camera_params.distort_coeffs, self.R_vec, self.T_vec, arm_length, 3, sub_pixel_fac)


def read_dict_from_file(fileName:str|pathlib.Path, episodes:list[list[int]]=None) -> dict[int,Pose]:
    return data_files.read_file(fileName,
                                Pose, True, True, False, False,
                                episodes=episodes)[0]

def write_list_to_file(poses: list[Pose], fileName:str|pathlib.Path, skip_failed=False):
    data_files.write_array_to_file(poses, fileName,
                                    Pose._columns_compressed,
                                    skip_all_nan=skip_failed)


# intervals are either lists of lists ([1,2], [100,200]), or dicts containing such lists as values
from . import annotation

def _prep_ival_dict(intervals, add_incomplete_intervals=False):
    intervals = intervals.copy()    # so any manipulation doesn't propagate back out
    for k in intervals:
        if not intervals[k]:
            continue
        if not isinstance(intervals[k][0], list):
            if not isinstance(k,annotation.Event) or annotation.type_map[k]==annotation.Type.Interval:
                temp = []
                for m in range(0,len(intervals[k])-1,2): # read in batches of two, and run until -1 to make sure we don't pick up incomplete intervals
                    temp.append(intervals[k][m:m+2])
                if add_incomplete_intervals and len(intervals[k])%2==1:
                    temp.append(intervals[k][-1:])
                intervals[k] = temp
            else:
                intervals[k] = [[tp] for tp in intervals[k]]
    return intervals

def is_in_interval(frame_idx, intervals):
    if intervals is None:
        return True # no interval defined, that means all frames should be processed

    # if its a dict, flatten it
    if isinstance(intervals, dict):
        intervals = _prep_ival_dict(intervals)
        intervals = [iv for k in intervals for iv in intervals[k]]

    # return True if we're in a current interval
    for iv in intervals:
        if len(iv)==1:
            if frame_idx==iv[0]:
                # exactly on the frame of a time point coding
                return True
        elif frame_idx>=iv[0] and frame_idx<=iv[1]:
            return True
    return False

def which_interval(frame_idx, intervals):
    if not isinstance(intervals, dict):
        return None, None
    # prep input, if needed
    intervals = _prep_ival_dict(intervals, add_incomplete_intervals=True)

    # get output
    keys = []
    ivals = []
    for k in intervals:
        if not intervals[k]:
            continue
        for iv in intervals[k]:
            if len(iv)==1:
                if frame_idx==iv[0]:
                    keys.append(k)
                    ivals.append(iv)
            elif frame_idx>=iv[0] and frame_idx<=iv[1]:
                keys.append(k)
                ivals.append(iv)

    return keys, ivals

def beyond_last_interval(frame_idx, intervals):
    if not intervals:
        return False
    elif isinstance(intervals, dict):
        for k in intervals:
            if not intervals[k]:
                return False
            if isinstance(intervals[k][-1], list):
                if frame_idx <= intervals[k][-1][-1]:
                    return False
            elif frame_idx <= intervals[k][-1]:
                return False
        return True
    else:
        return frame_idx > intervals[-1][-1]


""" Mpeg lookup tables """

mp4ra_table = {
    0x20: '0x20, mp4v',
    0x21: '0x21, avc1|avc2|avc3|avc4',
    0x22: '0x22, avcp',
    0x23: '0x23, hev1|hev2|hev3|hvc1|hvc2|hvc3',
    0x40: '0x40, m4ae|mp4a',
    0xA0: '0xA0, sevc',
    0xA1: '0xA1, ssmv',
    0xA3: '0xA3, vc-1',
    0xA4: '0xA4, drac',
    0xA5: '0xA5, ac-3',
    0xA6: '0xA6, ec-3',
    0xA7: '0xA7, dra1',
    0xA8: '0xA8, g719',
    0xA9: '0xA9, dtsc',
    0xAA: '0xAA, dtsh',
    0xAB: '0xAB, dtsl',
    0xAC: '0xAC, dtse',
    0xAD: '0xAD, Opus',
    0xAE: '0xAE, ac-4',
    0xB0: '0xB0, rv60',
    0xB1: '0xB1, vp09',
    0xB2: '0xB2, dtsx',
    0xB3: '0xB3, dtsy',
    0xE1: '0xE1, sqcp'
    }

es_table = {
    0x00: '0x00,Forbidden',
    0x01: '0x01,ObjectDescriptorStream (see 8.5)',
    0x02: '0x02,ClockReferenceStream (see 10.2.5)',
    0x03: '0x03,SceneDescriptionStream (see 9.2.1)',
    0x04: '0x04,VisualStream',
    0x05: '0x05,AudioStream',
    0x06: '0x06,MPEG7Stream',
    0x07: '0x07,IPMPStream (see 8.3.2)',
    0x08: '0x08,ObjectContentInfoStream (see 8.4.2)',
    0x09: '0x09,MPEGJStream',
    0x0A: '0x0A,Interaction Stream',
    0x0B: '0x0B,IPMPToolStream',
    0x0C: '0x0C,FontDataStream',
    0x0D: '0x0D,StreamingText'
    }

sound_codec_table = {
    0: '0, Null',
    1: '1, AAC Main',
    2: '2, AAC LC (Low Complexity)',
    3: '3, AAC SSR (Scalable Sample Rate)',
    4: '4, AAC LTP (Long Term Prediction)',
    5: '5, SBR (Spectral Band Replication)',
    6: '6, AAC Scalable',
    7: '7, TwinVQ',
    8: '8, CELP (Code Excited Linear Prediction)',
    9: '9, HXVC (Harmonic Vector eXcitation Coding)',
    10: '10, Reserved',
    11: '11, Reserved',
    12: '12, TTSI (Text-To-Speech Interface)',
    13: '13, Main Synthesis',
    14: '14, Wavetable Synthesis',
    15: '15, General MIDI',
    16: '16, Algorithmic Synthesis and Audio Effects',
    17: '17, ER (Error Resilient) AAC LC',
    18: '18, Reserved',
    19: '19, ER AAC LTP',
    20: '20, ER AAC Scalable',
    21: '21, ER TwinVQ',
    22: '22, ER BSAC (Bit-Sliced Arithmetic Coding)',
    23: '23, ER AAC LD (Low Delay)',
    24: '24, ER CELP',
    25: '25, ER HVXC',
    26: '26, ER HILN (Harmonic and Individual Lines plus Noise)',
    27: '27, ER Parametric',
    28: '28, SSC (SinuSoidal Coding)',
    29: '29, PS (Parametric Stereo)',
    30: '30, MPEG Surround',
    31: '31, (Escape value)',
    32: '32, Layer-1',
    33: '33, Layer-2',
    34: '34, Layer-3',
    35: '35, DST (Direct Stream Transfer)',
    36: '36, ALS (Audio Lossless)',
    37: '37, SLS (Scalable LosslesS)',
    38: '38, SLS non-core',
    39: '39, ER AAC ELD (Enhanced Low Delay)',
    40: '40, SMR (Symbolic Music Representation) Simple',
    41: '41, SMR Main',
    42: '42, USAC (Unified Speech and Audio Coding) (no SBR)',
    43: '43, SAOC (Spatial Audio Object Coding)',
    44: '44, LD MPEG Surround',
    45: '45, USAC',
    46: '46,Unknown',
    47: '47,Unknown',
    48: '48,Unknown',
    49: '49,Unknown',
    50: '50,Unknown',
    51: '51,Unknown',
    52: '52,Unknown',
    53: '53,Unknown',
    54: '54,Unknown',
    55: '55,Unknown',
    56: '56,Unknown',
    57: '57,Unknown',
    58: '58,Unknown',
    59: '59,Unknown',
    60: '60,Unknown',
    61: '61,Unknown',
    62: '62,Unknown',
    63: '63,Unknown'
    }

freq_table = {
    0: '0, 96000 Hz',
    1: '1, 88200 Hz',
    2: '2, 64000 Hz',
    3: '3, 48000 Hz',
    4: '4, 44100 Hz',
    5: '5, 32000 Hz',
    6: '6, 24000 Hz',
    7: '7, 22050 Hz',
    8: '8, 16000 Hz',
    9: '9, 12000 Hz',
    10: '10, 11025 Hz',
    11: '11, 8000 Hz',
    12: '12, 7350 Hz',
    13: '13, Reserved',
    14: '14, Reserved',
    15: '15, frequency is written explicitly'
    }

sound_channel_table = {
    0: '0, Defined in AOT Specifc Config',
    1: '1, 1 channel',
    2: '2, 2 channels',
    3: '3, 3 channels',
    4: '4, 4 channels',
    5: '5, 5 channels',
    6: '6, 6 channels',
    7: '7, 8 channels',
    8: '8,unknown',
    9: '9,unknown',
    10: '10,unknown',
    11: '11,unknown',
    12: '12,unknown',
    13: '13,unknown',
    14: '14,unknown',
    15: '15,unknown'
}



"""
iso.py

This file, if and when complete ;-), contains all the class definitions of boxes that are specified in ISO/IEC 14496-12.
Additionally a class to represent the MP4 file that contains the MP4 boxes has been defined.
A box_factory function has also been defined, primarily to minimise coupling between modules.

"""
import datetime
import logging

from . import non_iso
from .core import *
from .util import *
from .summary import *


# Supported box
# 'ftyp', 'pdin', 'moov', 'mvhd', 'meta', 'trak', 'tkhd', 'tref', 'trgr', 'edts', 'elst', 'mdia',
# 'mdhd', 'hdlr', 'elng', 'minf', 'vmhd', 'smhd', 'hmhd', 'nmhd', 'dinf', 'dref', 'url ', 'urn ',
# 'stbl', 'stsd', 'stts', 'ctts', 'cslg', 'stsc', 'stsz', 'stz2', 'stco', 'co64', 'stss', 'stsh',
# 'padb', 'stdp', 'sdtp', 'sbgp', 'sgpd', 'subs', 'saiz', 'saio', 'udta', 'mvex', 'mehd', 'trex',
# 'leva', 'moof', 'mfhd', 'traf', 'tfhd', 'trun', 'tfdt', 'mfra', 'tfra', 'mfro', 'mdat', 'free',
# 'skip', 'cprt', 'tsel', 'strk', 'stri', 'strd', 'iloc', 'ipro', 'rinf', 'sinf', 'frma', 'schm',
# 'xml ', 'pitm', 'iref', 'meco', 'mere', 'styp', 'sidx', 'ssix', 'prft', 'avc1', 'hvc1', 'avcC',
# 'hvcC', 'btrt', 'pasp', 'mp4a', 'ac-3', 'ec-3', 'esds', 'dac3', 'dec3', 'ilst', 'data', 'pssh',
# 'senc'
# Not supported
# 'sthd', 'iinf', 'bxml', 'fiin', 'paen', 'fire', 'fpar', 'fecr', 'segr', 'gitn', 'idat'

def box_factory(fp, header, parent):
    """
    box_factory() takes a box type as a parameter and, if a class has been defined for that type, returns
    an instance of that class.
    """
    the_box = None
    # Normalise header type so it can be expressed in a Python Class name
    box_type = header.type.replace(" ", "_").replace("-", "_").lower()
    _box_class = globals().get(
        box_type.capitalize() + 'Box')  # globals() Return a dictionary representing the current global symbol table
    if _box_class:
        the_box = _box_class(fp, header, parent)
        return the_box
    else:
        return non_iso.box_factory_non_iso(fp, header, parent)


class Mp4File:
    """ Mp4File Class, effectively the top-level container """
    def __init__(self, filename):
        self.filename = filename
        self.type = 'file'
        self.children = []
        self.summary= {}
        try:
            with open(filename, 'rb') as f:
                end_of_file = False
                while not end_of_file:
                    current_header = Header(f)
                    current_box = box_factory(f, current_header, self)
                    self.children.append(current_box)
                    if current_box.size == 0:
                        end_of_file = True
                    if len(f.read(4)) != 4:
                        end_of_file = True
                    else:
                        f.seek(-4, 1)
            f.close()
            self._generate_samples_from_moov()
            self._generate_samples_from_moofs()
        except Exception as e:
            # catch exception in case we can continue
            logging.exception(f'error in {filename} after child {len(self.children)}')

    def _generate_samples_from_moov(self):
        """ identify media samples in mdat for full mp4 file """
        mdats = [mbox for mbox in self.children if mbox.type == 'mdat']
        # generate a sample list if there is a moov that contains traks N.B only ever 0,1 moov boxes
        if [box for box in self.children if box.type == 'moov']:
            moov = [box for box in self.children if box.type == 'moov'][0]
            traks = [tbox for tbox in moov.children if tbox.type == 'trak']
            sample_list = []
            for trak in traks:
                trak_id = [box for box in trak.children if box.type == 'tkhd'][0].box_info['track_ID']
                timescale = [box for box in [box for box in trak.children
                                             if box.type == 'mdia'][0].children
                             if box.type == 'mdhd'][0].box_info['timescale']
                samplebox = [box for box in [box for box in [box for box in trak.children
                                                             if box.type == 'mdia'][0].children
                                             if box.type == 'minf'][0].children
                             if box.type == 'stbl'][0]
                chunk_offsets = [box for box in samplebox.children
                                 if box.type == 'stco' or box.type == 'co64'][0].box_info['entry_list']
                sample_size_box = [box for box in samplebox.children if box.type == 'stsz' or box.type == 'stz2'][0]
                if sample_size_box.box_info['sample_size'] > 0:
                    sample_sizes = [{'entry_size': sample_size_box.box_info['sample_size']}
                                    for i in range(sample_size_box.box_info['sample_count'])]
                else:
                    sample_sizes = sample_size_box.box_info['entry_list']
                sample_to_chunks = [box for box in samplebox.children
                                    if box.type == 'stsc'][0].box_info['entry_list']
                s2c_index = 0
                next_run = 0
                sample_idx = 0
                for i, chunk in enumerate(chunk_offsets, 1):
                    if i >= next_run:
                        samples_per_chunk = sample_to_chunks[s2c_index]['samples_per_chunk']
                        s2c_index += 1
                        next_run = sample_to_chunks[s2c_index]['first_chunk'] \
                            if s2c_index < len(sample_to_chunks) else len(chunk_offsets) + 1
                    chunk_dict = {'track_ID': trak_id,
                                  'chunk_ID': i,
                                  'chunk_offset': chunk['chunk_offset'],
                                  'samples_per_chunk': samples_per_chunk,
                                  'chunk_samples': []
                                  }
                    sample_offset = chunk['chunk_offset']
                    for j, sample in enumerate(sample_sizes[sample_idx:sample_idx + samples_per_chunk], sample_idx + 1):
                        chunk_dict['chunk_samples'].append({
                            'sample_ID': j,
                            'size': sample['entry_size'],
                            'offset': sample_offset
                        })
                        sample_offset += sample['entry_size']
                    sample_list.append(chunk_dict)
                    sample_idx += samples_per_chunk
            # sample_list could be empty, say, for mpeg-dash initialization segment
            if sample_list:
                # sort by chunk offset to get interleaved list
                sample_list.sort(key=lambda k: k['chunk_offset'])
                for mdat in mdats:
                    mdat_sample_list = [sample for sample in sample_list if
                                        mdat.start_of_box < sample['chunk_offset'] < (mdat.start_of_box + mdat.size)]
                    if len(mdat_sample_list):
                        mdat.box_info['message'] = 'Has samples.'
                    mdat.sample_list = mdat_sample_list

    def _generate_samples_from_moofs(self):
        """
        generate samples within mdats of media segments for fragmented mp4 files
        media segments are 1 moof (optionally preceded by an styp) followed by 1 or more contiguous mdats
        I've only ever seen 1 mdat in a media segment though
        """
        i = 0
        while i < len(self.children) - 1:
            if self.children[i].type == 'moof':
                moof = self.children[i]
                media_segment = {'moof_box': moof, 'mdat_boxes': []}
                sequence_number = [mfhd for mfhd in moof.children
                                   if mfhd.type == 'mfhd'][0].box_info['sequence_number']
                while i < len(self.children) - 1 and self.children[i + 1].type == 'mdat':
                    media_segment['mdat_boxes'].append(self.children[i + 1])
                    i += 1
                # I've only ever seen 1 traf in a moof, but the standard says there could be more
                data_offset = 0
                for j,traf in enumerate([tbox for tbox in moof.children if tbox.type == 'traf']):
                    # read tfhd, there will be one
                    tfhd = [hbox for hbox in traf.children if hbox.type == 'tfhd'][0]
                    trak_id = tfhd.box_info['track_id']
                    if 'base_data_offset' in tfhd.box_info:
                        data_offset = tfhd.box_info['base_data_offset']
                    elif tfhd.box_info['default_base_is_moof']:
                        data_offset = media_segment['moof_box'].start_of_box
                    elif j > 0:
                        # according to spec. should be set end of data for last fragment
                        pass
                    else:
                        base_data_offset = media_segment['moof_box'].start_of_box
                    for k, trun in enumerate([rbox for rbox in traf.children if rbox.type == 'trun'], 1):
                        if 'data_offset' in trun.box_info:
                            data_offset += trun.box_info['data_offset']
                        run_dict = {'sequence_number': sequence_number,
                                    'track_ID': trak_id,
                                    'run_ID': k,
                                    'run_offset': data_offset,
                                    'sample_count': trun.box_info['sample_count'],
                                    'run_samples': []
                                    }
                        has_sample_size = True if trun.flags & 0x0200 == 0x0200 else False
                        for l, sample in enumerate(trun.box_info['samples'], 1):
                            if not has_sample_size:
                                sample_size = tfhd.box_info['default_sample_size']
                            else:
                                sample_size = sample['sample_size']
                            run_dict['run_samples'].append({'sample_ID': l,
                                                            'size': sample_size,
                                                            'offset': data_offset
                                                            })
                            data_offset += sample_size
                        for mdat in media_segment['mdat_boxes']:
                            if mdat.start_of_box < run_dict['run_offset'] and (mdat.start_of_box
                                                                               + mdat.size) >= data_offset:
                                mdat.box_info['message'] = 'Has samples.'
                                mdat.sample_list.append(run_dict)
            i += 1

    def read_bytes(self, offset, num_bytes):
        with open(self.filename, 'rb') as f:
            f.seek(offset)
            bytes_read = f.read(num_bytes)
        f.close()
        return bytes_read

    def get_summary(self):
        if not self.summary:
            self.summary = Summary(self)
        return self.summary.data

    def search_boxes_for_type(self, box_type):
        type_matches = []
        for box in self.children:
            if box.type == box_type:
                type_matches.append(box)
            if box.children:
                type_matches += box.search_child_boxes_for_type(box_type)
        return type_matches
# Box classes


class FreeBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            pass
        finally:
            fp.seek(self.start_of_box + self.size)


SkipBox = FreeBox


class FtypBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info = {
                'major_brand': fp.read(4).decode('utf-8'),
                'minor_version': "{0:#010x}".format(read_u32(fp)),
                'compatible_brands': []
            }
            bytes_left = self.size - (self.header.header_size + 8)
            while bytes_left > 0:
                self.box_info['compatible_brands'].append(fp.read(4).decode('utf-8'))
                bytes_left -= 4
        finally:
            fp.seek(self.start_of_box + self.size)


StypBox = FtypBox


class ColrBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['color_type'] = fp.read(4).decode('utf-8')
            if self.box_info['color_type'] == 'nclx':
                self.box_info['color_primaries'] = read_u16(fp)
                self.box_info['transfer_characteristics'] = read_u16(fp)
                self.box_info['matrix_coefficients'] = read_u16(fp)
                self.box_info['full_range_flag'] = read_u8(fp) >> 7
        finally:
            fp.seek(self.start_of_box + self.size)


class PdinBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        end_of_box = self.start_of_box + self.size
        try:
            self.box_info['rates'] = []
            while fp.tell() < end_of_box:
                self.box_info('rates').append({'rate': read_u32(fp), 'initial_delay': read_u32(fp)})
        finally:
            fp.seek(end_of_box)


class ContainerBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            bytes_left = self.size - self.header.header_size
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


# All these are pure container boxes
DinfBox = MinfBox = MdiaBox = TrefBox = EdtsBox = TrakBox = MoofBox = MoovBox = ContainerBox
UdtaBox = TrgrBox = MvexBox = MfraBox = StrkBox = StrdBox = RinfBox = SinfBox = MecoBox = ContainerBox
GmhdBox = SchiBox = ContainerBox


class MetaBox(Mp4Box):
    """
    Seems to be a discrepancy between Apple atom spec and ISO about whether this is a versioned box
    """
    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            bytes_left = self.size - self.header.header_size
            first_four_bytes = read_u32(fp)
            second_four_bytes = fp.read(4)
            if second_four_bytes.decode('utf-8', errors="ignore") == 'hdlr':
                # it's non-versioned
                fp.seek(-8, 1)
            else:
                # it's versioned
                self.version = first_four_bytes >> 24
                self.flags = first_four_bytes & 0xFFFFFF
                fp.seek(-4, 1)
                bytes_left -= 4
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


class MdatBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        self.sample_list = []
        try:
            self.box_info['message'] = 'No samples found.'
        finally:
            fp.seek(self.start_of_box + self.size)


class MvhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            dt_base = datetime.datetime(1904, 1, 1, 0, 0, 0)
            if self.version == 1:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['timescale'] = read_u32(fp)
                self.box_info['duration'] = read_u64(fp)
            else:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['timescale'] = read_u32(fp)
                self.box_info['duration'] = read_u32(fp)
            self.box_info['rate'] = read_u16_16(fp)
            self.box_info['volume'] = read_u8_8(fp)
            fp.seek(10, 1)
            self.box_info['matrix'] = ["{0:#010x}".format(b) for b in struct.unpack('>9I', fp.read(36))]
            fp.seek(24, 1)
            self.box_info['next_track_id'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class MfhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['sequence_number'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class MehdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.version == 1:
                self.box_info['fragment_duration'] = read_u64(fp)
            else:
                self.box_info['fragment_duration'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class ElstBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                if self.version == 1:
                    self.box_info['entry_list'].append({
                        'segment_duration': read_u64(fp),
                        'media_time': read_i64(fp),
                        'media_rate': read_u16_16(fp),
                    })
                else:
                    self.box_info['entry_list'].append({
                        'segment_duration': read_u32(fp),
                        'media_time': read_i32(fp),
                        'media_rate': read_u16_16(fp),
                    })
        finally:
            fp.seek(self.start_of_box + self.size)


class TkhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            dt_base = datetime.datetime(1904, 1, 1, 0, 0, 0)
            if self.version == 1:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['track_ID'] = read_u32(fp)
                fp.seek(4, 1)
                self.box_info['duration'] = read_u64(fp)
            else:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['track_ID'] = read_u32(fp)
                fp.seek(4, 1)
                self.box_info['duration'] = read_u32(fp)
            fp.seek(8, 1)
            self.box_info['layer'] = read_i16(fp)
            self.box_info['alternate_group'] = read_i16(fp)
            self.box_info['volume'] = read_u8_8(fp)
            fp.seek(2, 1)
            self.box_info['matrix'] = ["{0:#010x}".format(b) for b in struct.unpack('>9I', fp.read(36))]
            self.box_info['width'] = read_u16_16(fp)
            self.box_info['height'] = read_u16_16(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class TfhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['track_id'] = read_u32(fp)
            if self.flags  & 0x000001 == 0x000001:
                self.box_info['base_data_offset'] = read_u64(fp)
            if self.flags  & 0x000002 == 0x000002:
                self.box_info['sample_description_index'] = read_u32(fp)
            if self.flags  & 0x000008 == 0x000008:
                self.box_info['default_sample_duration'] = read_u32(fp)
            if self.flags  & 0x000010 == 0x000010:
                self.box_info['default_sample_size'] = read_u32(fp)
            if self.flags  & 0x000020 == 0x000020:
                self.box_info['default_sample_flags'] = "{0:#08x}".format(read_u32(fp))
            self.box_info['duration_is_empty'] = True if self.flags  & 0x010000 == 0x010000 else False
            # depending on value of base data offset this flag mght be ignored
            self.box_info['default_base_is_moof'] = True if self.flags  & 0x020000 == 0x020000 else False
        finally:
            fp.seek(self.start_of_box + self.size)


class TrexBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['track_ID'] = read_u32(fp)
            self.box_info['default_sample_description_index'] = read_u32(fp)
            self.box_info['default_sample_duration'] = read_u32(fp)
            self.box_info['default_sample_size'] = read_u32(fp)
            self.box_info['default_sample_flags'] = "{0:#08x}".format(read_u32(fp))
        finally:
            fp.seek(self.start_of_box + self.size)


class LevaBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['level_count'] = read_u8(fp)
            self.box_info['level_list'] = []
            for i in range(self.box_info['level_count']):
                level_dict = {'track_ID': read_u32(fp)}
                pad_assign = read_u8(fp)
                level_dict['padding_flag'] = pad_assign // 128
                level_dict['assignment_type'] = pad_assign % 128
                if level_dict['assignment_type'] == 0:
                    level_dict['grouping_type'] = fp.read(4).decode('utf-8')
                elif level_dict['assignment_type'] == 1:
                    level_dict['grouping_type'] = fp.read(4).decode('utf-8')
                    level_dict['grouping_type_parameter'] = read_u32(fp)
                elif level_dict['assignment_type'] == 4:
                    level_dict['sub_track_id'] = read_u32(fp)
                self.box_info['level_list'].append(level_dict)
        finally:
            fp.seek(self.start_of_box + self.size)


class TfraBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['track_ID'] = read_u32(fp)
            length_fields = read_u32(fp)
            self.box_info['length_size_of_traf_num'] = length_fields >> 4 & 3
            self.box_info['length_size_of_trun_num'] = length_fields >> 2 & 3
            self.box_info['length_size_of_sample_num'] = length_fields & 3
            self.box_info['number_of_entry'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['number_of_entry']):
                entry_dict = {}
                if self.version == 1:
                    entry_dict['time'] = read_u64(fp)
                    entry_dict['moof_offset'] = read_u64(fp)
                else:
                    entry_dict['time'] = read_u32(fp)
                    entry_dict['moof_offset'] = read_u32(fp)
                if self.box_info['length_size_of_traf_num'] == 0:
                    entry_dict['traf_number'] = read_u8(fp)
                elif self.box_info['length_size_of_traf_num'] == 1:
                    entry_dict['traf_number'] = read_u16(fp)
                elif self.box_info['length_size_of_traf_num'] == 3:
                    entry_dict['traf_number'] = read_u32(fp)
                if self.box_info['length_size_of_trun_num'] == 0:
                    entry_dict['trun_number'] = read_u8(fp)
                elif self.box_info['length_size_of_trun_num'] == 1:
                    entry_dict['trun_number'] = read_u16(fp)
                elif self.box_info['length_size_of_trun_num'] == 3:
                    entry_dict['trun_number'] = read_u32(fp)
                if self.box_info['length_size_of_sample_num'] == 0:
                    entry_dict['sample_number'] = read_u8(fp)
                elif self.box_info['length_size_of_sample_num'] == 1:
                    entry_dict['sample_number'] = read_u16(fp)
                elif self.box_info['length_size_of_sample_num'] == 3:
                    entry_dict['sample_number'] = read_u32(fp)
                self.box_info['entry_list'].append(entry_dict)
        finally:
            fp.seek(self.start_of_box + self.size)


class MfroBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['size'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class CprtBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            # I think this is right
            lang = read_u16(fp)
            if lang == 0:
                self.box_info['language'] = '0x00'
            else:
                ch1 = str(chr(96 + (lang >> 10 & 31)))
                ch2 = str(chr(96 + (lang >> 5 & 31)))
                ch3 = str(chr(96 + (lang & 31)))
                self.box_info['language'] = ch1 + ch2 + ch3
            bytes_left = self.size - (self.header.header_size + 2)
            self.box_info['name'] = fp.read(bytes_left).decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)


class TselBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['switch_group'] = read_u32(fp)
            bytes_left = self.size - (self.header.header_size + 8)
            attr_list = []
            while bytes_left > 0:
                attr_list.append(fp.read(4).decode('utf-8'))
                bytes_left -= 4
            self.box_info['attributes'] = attr_list
        finally:
            fp.seek(self.start_of_box + self.size)


class StriBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['switch_group'] = read_u16(fp)
            self.box_info['alternate_group'] = read_u16(fp)
            self.box_info['sub_track_ID'] = read_u32(fp)
            bytes_left = self.size - (self.header.header_size + 12)
            attr_list = []
            while bytes_left > 0:
                attr_list.append(fp.read(4).decode('utf-8'))
                bytes_left -= 4
            self.box_info['attributes'] = attr_list
        finally:
            fp.seek(self.start_of_box + self.size)


class IlocBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['offset_size'] = read_u32(fp)
            self.box_info['length_size'] = read_u32(fp)
            self.box_info['base_offset_size'] = read_u32(fp)
            if self.version == 1 or self.version == 2:
                self.box_info['index_size'] = read_u32(fp)
            else:
                self.box_info['reserved'] = read_u32(fp)
            if self.version < 2:
                self.box_info['item_count'] = read_u16(fp)
            elif self.version == 2:
                self.box_info['item_count'] = read_u32(fp)
            self.box_info['item_list'] = []
            for i in range(self.box_info['item_count']):
                item = {}
                if self.version < 2:
                    item['item_ID'] = read_u16(fp)
                elif self.version == 2:
                    item['item_ID'] = read_u32(fp)
                if self.version == 1 or self.version == 2:
                    item['construction_method'] = read_u16(fp) % 16
                item['data_reference_index'] = read_u16(fp)
                if self.box_info['offset_size'] == 4:
                    item['base_offset'] = read_u32(fp)
                elif self.box_info['offset_size'] == 8:
                    item['base_offset'] = read_u64(fp)
                item['extent_count'] = read_u16(fp)
                item['extent_list'] = []
                for j in range(item['extent_count']):
                    extent = {}
                    if self.version == 1 or self.version == 2:
                        if self.box_info['index_size'] == 4:
                            extent['extent_index'] = read_u32(fp)
                        elif self.box_info['index_size'] == 8:
                            extent['extent_index'] = read_u64(fp)
                    if self.box_info['offset_size'] == 4:
                        extent['extent_offset'] = read_u32(fp)
                    elif self.box_info['offset_size'] == 8:
                        extent['extent_offset'] = read_u64(fp)
                    if self.box_info['length_size'] == 4:
                        extent['extent_length'] = read_u32(fp)
                    elif self.box_info['length_size'] == 8:
                        extent['extent_length'] = read_u64(fp)
                    item['extent_list'].append(extent)
                self.box_info['item_list'].append(item)
        finally:
            fp.seek(self.start_of_box + self.size)


class IproBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['protection_count'] = read_u16(fp)
            for i in range(self.box_info['protection_count']):
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
        finally:
            fp.seek(self.start_of_box + self.size)


class FrmaBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['data_format'] = fp.read(4).decode('utf-8')
        finally:
            fp.seek(self.start_of_box + self.size)


class SchmBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['scheme_type'] = fp.read(4).decode('utf-8')
            self.box_info['scheme_version'] = read_u32(fp)
            if self.flags & 0x000001 == 0x000001:
                self.box_info['data_offset'] = fp.read(self.size - (self.header.header_size + 12)).decode('utf-8')
        finally:
            fp.seek(self.start_of_box + self.size)


class Xml_Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['xml_data'] = fp.read(self.size - (self.header.header_size + 4)).decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)


class PitmBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.version == 0:
                self.box_info['item_ID'] = read_u16(fp)
            else:
                self.box_info['item_ID'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


# This is just a versioned container box
class IrefBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            bytes_left = self.size - self.header.header_size
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


class MereBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['first_metabox_handler_type'] = fp.read(4).decode('utf-8')
            self.box_info['second_metabox_handler_type'] = fp.read(4).decode('utf-8')
            self.box_info['metabox_relation'] = read_u8(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class TrunBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['sample_count'] = read_u32(fp)
            if self.flags & 0x000001 == 0x000001:
                self.box_info['data_offset'] = read_i32(fp)
            if self.flags & 0x000004 == 0x000004:
                self.box_info['first_sample_flags'] = "{0:#08x}".format(read_u32(fp))
            has_sample_duration = True if self.flags & 0x000100 == 0x000100 else False
            has_sample_size = True if self.flags & 0x000200 == 0x000200 else False
            has_sample_flags = True if self.flags & 0x000400 == 0x000400 else False
            has_scto = True if self.flags & 0x000800 == 0x000800 else False
            sample_list = []
            for i in range(self.box_info['sample_count']):
                sample = {}
                if has_sample_duration:
                    sample['sample_duration'] = read_u32(fp)
                if has_sample_size:
                    sample['sample_size'] = read_u32(fp)
                if has_sample_flags:
                    sample['sample_flags'] = "{0:#08x}".format(read_u32(fp))
                if has_scto:
                    if self.version == 1:
                        self.box_info['sample_composition_time_offset'] = read_i32(fp)
                    else:
                        self.box_info['sample_composition_time_offset'] = read_u32(fp)
                sample_list.append(sample)
            self.box_info['samples'] = sample_list
        finally:
            fp.seek(self.start_of_box + self.size)


class TfdtBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.version == 1:
                self.box_info['baseMediaDecode'] = read_u64(fp)
            else:
                self.box_info['baseMediaDecode'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class MdhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            dt_base = datetime.datetime(1904, 1, 1, 0, 0, 0)
            if self.version == 1:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u64(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['timescale'] = read_u32(fp)
                self.box_info['duration'] = read_u64(fp)
            else:
                self.box_info['creation_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['modification_time'] = (
                        dt_base + datetime.timedelta(seconds=(read_u32(fp)))).strftime('%Y-%m-%d %H:%M:%S')
                self.box_info['timescale'] = read_u32(fp)
                self.box_info['duration'] = read_u32(fp)
            # I think this is right
            lang = struct.unpack('>H', fp.read(2))[0]
            if lang == 0:
                self.box_info['language'] = '0x00'
            else:
                ch1 = str(chr(96 + (lang >> 10 & 31)))
                ch2 = str(chr(96 + (lang >> 5 & 31)))
                ch3 = str(chr(96 + (lang % 32)))
                self.box_info['language'] = ch1 + ch2 + ch3
        finally:
            fp.seek(self.start_of_box + self.size)


class ElngBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['extended_language'] = fp.read(self.size - (self.header.header_size + 4)).split(b'\x00')[0]
        finally:
            fp.seek(self.start_of_box + self.size)


class DrefBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            for i in range(self.box_info['entry_count']):
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
        finally:
            fp.seek(self.start_of_box + self.size)


class Url_Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.flags & 0x000001 != 1:
                data_entry = fp.read(self.size - (self.header.header_size + 4))
                self.box_info['location'] = data_entry.decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)


class Urn_Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.flags & 0x000001 != 1:
                name, ignore, location = fp.read(self.size - (self.header.header_size + 4)).partition(b'\x00')
                self.box_info['name'] = location.decode('utf-8', errors="ignore")
                self.box_info['location'] = location.decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)


class HdlrBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            fp.seek(4, 1)
            self.box_info['handler_type'] = fp.read(4).decode('utf-8')
            fp.seek(12, 1)
            bytes_left = self.size - (self.header.header_size + 25)  # string is null terminated
            self.box_info['name'] = fp.read(bytes_left).decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)


class StblBox(ContainerBox):
    # Sub-class from container box so we can do some extra things with child boxes
    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            # Some sample table boxes have dependencies on other sample table table boxes in order to read correctly
            # Fill stdp, sdtp lists using sample count in stsz
            sz = [box for box in self.children if box.type == 'stsz' or box.type == 'stz2'][0] if [
                box for box in self.children if box.type == 'stsz' or box.type == 'stz2'] else False
            sc = sz.box_info['sample_count'] if sz else False
            sdtp = [box for box in self.children if box.type == 'sdtp'][0] if [
                box for box in self.children if box.type == 'sdtp'] else False
            stdp = [box for box in self.children if box.type == 'stdp'][0] if [
                box for box in self.children if box.type == 'stdp'] else False
            if sc and sdtp:
                sdtp.update_table(fp, sc)
            if sc and stdp:
                stdp.update_table(fp, sc)
        finally:
            fp.seek(self.start_of_box + self.size)


class TrafBox(ContainerBox):
    # Sub-class from container box so we can do some extra things with child boxes
    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            # if we have a senc box
            senc = [box for box in self.children if box.type == 'senc'][0] if [
                box for box in self.children if box.type == 'senc'] else False
            if senc:
                # check if it has sub-sampling before getting IV_size from sgpd or saiz
                if senc.flags & 0x000002 == 0x000002:
                    per_sample_iv_size = 0
                    sgpd = [box for box in self.children if box.type == 'sgpd'][0] if [
                       box for box in self.children if box.type == 'sgpd'] else False
                    saiz = [box for box in self.children if box.type == 'saiz'][0] if [
                        box for box in self.children if box.type == 'saiz'] else False
                    if sgpd and sgpd.box_info['grouping_type'] == 'seig':
                        # I've only ever seen a single entry in seig so get IV size from this
                        per_sample_iv_size = sgpd.box_info['entry_list'][0]['per_sample_iv_size']
                    # try saiz
                    elif saiz:
                        # does it have non-zero default size?
                        if saiz.box_info['default_sample_info_size'] > 0:
                            sample_size = saiz.box_info['default_sample_info_size']
                        else:
                            sample_size = min([sample['sample_info_size']
                                                for sample in saiz.box_info['sample_info_size_list']])
                        # deduct 10 or 18 from sample size (8 or 16 byte IV + 2 byte sub-sample count) and divide by six
                        # (2 bytes clear data size + 4 bytes enc data size)
                        per_sample_iv_size = 8 if (sample_size - 10) % 6 == 0 else 16 if (sample_size - 18) % 6 == 0 else 0
                    senc.populate_sample_table(fp, per_sample_iv_size)
        finally:
            fp.seek(self.start_of_box + self.size)


class VmhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['graphicsmode'] = read_u16(fp)
            self.box_info['opcolor'] = struct.unpack('>3H', fp.read(6))
        finally:
            fp.seek(self.start_of_box + self.size)


class SmhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['balance'] = read_i8_8(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class HmhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['maxPDUsize'] = read_u16(fp)
            self.box_info['avgPDUsize'] = read_u16(fp)
            self.box_info['maxbitrate'] = read_u32(fp)
            self.box_info['avgbitrate'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class NmhdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)

        try:
            pass
        finally:
            fp.seek(self.start_of_box + self.size)


class StsdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            for i in range(self.box_info['entry_count']):
                current_header = Header(fp)
                current_box = box_factory(fp, current_header, self)
                self.children.append(current_box)
        finally:
            fp.seek(self.start_of_box + self.size)


class SttsBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({'sample_count': read_u32(fp), 'sample_delta': read_u32(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class CttsBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                if self.version == 1:
                    self.box_info['entry_list'].append({'sample_count': read_u32(fp), 'sample_offset': read_i32(fp)})
                else:
                    self.box_info['entry_list'].append({'sample_count': read_u32(fp), 'sample_offset': read_i32(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class CslgBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.version == 1:
                self.box_info['compositionToDTSShift'] = read_i64(fp)
                self.box_info['leastDecodeToDisplayDelta'] = read_i64(fp)
                self.box_info['greatestDecodeToDisplayDelta'] = read_i64(fp)
                self.box_info['compositionStartTime'] = read_i64(fp)
                self.box_info['compositionEndTime'] = read_i64(fp)
            else:
                self.box_info['compositionToDTSShift'] = read_i32(fp)
                self.box_info['leastDecodeToDisplayDelta'] = read_i32(fp)
                self.box_info['greatestDecodeToDisplayDelta'] = read_i32(fp)
                self.box_info['compositionStartTime'] = read_i32(fp)
                self.box_info['compositionEndTime'] = read_i32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class StssBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({'sample_number': read_u32(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class StshBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({
                    'shadowed_sample_number': read_u32(fp),
                    'sync_sample_number': read_u32(fp)
                })
        finally:
            fp.seek(self.start_of_box + self.size)


class StscBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                first_chunk = read_u32(fp)
                samples_per_chunk = read_u32(fp)
                samples_description_index = read_u32(fp)
                self.box_info['entry_list'].append({'first_chunk': first_chunk, 'samples_per_chunk': samples_per_chunk,
                                                    'samples_description_index': samples_description_index})
        finally:
            fp.seek(self.start_of_box + self.size)


class StcoBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({'chunk_offset': read_u32(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class Co64Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({'chunk_offset': read_u64(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class PadbBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['sample_count'] = read_u32(fp)
            self.box_info['sample_list'] = []
            for i in range(self.box_info['sample_count']):
                pads = read_u8(fp)
                self.box_info['entry_list'].append({'pad1': pads // 16, 'pad2': pads % 16})
        finally:
            fp.seek(self.start_of_box + self.size)


class SubsBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                sample_delta = read_u32(fp)
                subsample_count = read_u16(fp)
                if subsample_count > 0:
                    subsample_list = []
                    for j in range(subsample_count):
                        if self.version == 1:
                            subsample_size = read_u32(fp)
                        else:
                            subsample_size = read_u16(fp)
                        subsample_priority = read_u8(fp)
                        discardable = read_u8(fp)
                        codec_specific_parameters = read_u32(fp)
                        subsample_list.append({
                            'subsample_size': subsample_size,
                            'subsample_priority': subsample_priority,
                            'discardable': discardable,
                            'codec_specific_parameters': codec_specific_parameters
                        })
                    self.box_info['entry_list'].append({
                        'sample_delta': sample_delta,
                        'subsample_count': subsample_count,
                        'subsample_list': subsample_list
                    })
        finally:
            fp.seek(self.start_of_box + self.size)


class SbgpBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['grouping_type'] = fp.read(4).decode('utf-8')
            if self.version == 1:
                self.box_info['grouping_type_parameter'] = read_u32(fp)
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                self.box_info['entry_list'].append({
                    'sample_count': read_u32(fp),
                    'group_description_index': read_u32(fp)
                })
        finally:
            fp.seek(self.start_of_box + self.size)


class SgpdBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['grouping_type'] = fp.read(4).decode('utf-8')
            if self.version == 1:
                self.box_info['default_length'] = read_u32(fp)
            elif self.version >= 2:
                self.box_info['default_sample_description_index'] = read_u32(fp)
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                if self.box_info['default_length'] == 0 and self.version == 1:
                    description_length = read_u32(fp)
                else:
                    description_length = self.box_info['default_length']
                sample_group_entry = binascii.b2a_hex(fp.read(description_length)).decode('utf-8')
                if self.box_info['grouping_type'] == 'seig':
                    seig_dict = {'crypto_byte_block': int(sample_group_entry[2], 16),
                                                            'skip_byte_block': int(sample_group_entry[3], 16),
                                                            'is_encrypted': int(sample_group_entry[5], 16),
                                                            'per_sample_iv_size': int(sample_group_entry[6:8], 16),
                                                            'kid': sample_group_entry[8:40]}
                    if seig_dict['is_encrypted'] == 1 and seig_dict['per_sample_iv_size'] == 0:
                        seig_dict['constant_IV_size'] = int(sample_group_entry[40:42], 16)
                        seig_dict['constant_IV'] = sample_group_entry[42:]
                    self.box_info['entry_list'].append(seig_dict)
                else:
                    self.box_info['entry_list'].append({'sample_group_entry': sample_group_entry})
        finally:
            fp.seek(self.start_of_box + self.size)


class SaizBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.flags & 0x000001 == 0x000001:
                self.box_info['aux_info_type'] = fp.read(4).decode('utf-8')
                self.box_info['aux_info_type_parameter'] = read_u32(fp)
            self.box_info['default_sample_info_size'] = read_u8(fp)
            self.box_info['sample_count'] = read_u32(fp)
            if self.box_info['default_sample_info_size'] == 0:
                self.box_info['sample_info_size_list'] = []
                for i in range(self.box_info['sample_count']):
                    self.box_info['sample_info_size_list'].append({'sample_info_size': read_u8(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class SaioBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            if self.flags & 0x000001 == 0x000001:
                self.box_info['aux_info_type'] = fp.read(4).decode('utf-8')
                self.box_info['aux_info_type_parameter'] = read_u32(fp)
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['offset_list'] = []
            for i in range(self.box_info['entry_count']):
                if self.version == 0:
                    self.box_info['offset_list'].append({'offset': read_u32(fp)})
                else:
                    self.box_info['offset_list'].append({'offset': read_u64(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class StszBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['sample_size'] = read_u32(fp)
            self.box_info['sample_count'] = read_u32(fp)
            if self.box_info['sample_size'] == 0:
                self.box_info['entry_list'] = []
                for i in range(self.box_info['sample_count']):
                    self.box_info['entry_list'].append({'entry_size': read_u32(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class Stz2Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['field_size'] = read_u32(fp)
            self.box_info['sample_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['sample_count']):
                if self.box_info['field_size'] == 4:
                    mybyte = read_u8(fp)
                    self.box_info['entry_list'].append({'entry_size': mybyte // 16}, {'entry_size+': mybyte % 16})
                if self.box_info['field_size'] == 8:
                    self.box_info['entry_list'].append({'entry_size': read_u8(fp)})
                if self.box_info['field_size'] == 16:
                    self.box_info['entry_list'].append({'entry_size': read_u16(fp)})
        finally:
            fp.seek(self.start_of_box + self.size)


class StdpBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            pass
        finally:
            fp.seek(self.start_of_box + self.size)

    def update_table(self, fp, sc):
        fp_orig = fp.tell()
        fp.seek(self.start_of_box + self.header.header_size + 4)
        self.box_info['sample_list'] = []
        for i in range(sc):
            self.box_info['sample_list'].append({'priority': read_u16(fp)})
        fp.seek(fp_orig)


class SdtpBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            pass
        finally:
            fp.seek(self.start_of_box + self.size)

    def update_table(self, fp, sc):
        fp_orig = fp.tell()
        fp.seek(self.start_of_box + self.header.header_size + 4)
        self.box_info['sample_list'] = []
        for i in range(sc):
            the_byte = read_u8(fp)
            is_leading = the_byte >> 6
            depends_on = the_byte >> 4 & 3
            is_depended_on = the_byte >> 2 & 3
            has_redundancy = the_byte & 3
            self.box_info['sample_list'].append({
                'is_leading': is_leading,
                'sample_depends_on': depends_on,
                'sample_is_depended_on': is_depended_on,
                'sample_has_redundancy': has_redundancy
            })
        fp.seek(fp_orig)


class SidxBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['reference_ID'] = read_u32(fp)
            self.box_info['timescale'] = read_u32(fp)
            if self.version == 0:
                self.box_info['earliest_presentation_time'] = read_u32(fp)
                self.box_info['first_offset'] = read_u32(fp)
            else:
                self.box_info['earliest_presentation_time'] = read_u64(fp)
                self.box_info['first_offset'] = read_u64(fp)
            fp.seek(2, 1)
            self.box_info['reference_count'] = read_u16(fp)
            self.box_info['reference_list'] = []
            for i in range(self.box_info['reference_count']):
                rt_sz = read_u32(fp)
                subsegment_dur = read_u32(fp)
                st_sz = read_u32(fp)
                self.box_info['reference_list'].append({
                    'reference_type': rt_sz >> 31,
                    'reference_size': rt_sz % 2147483648,
                    'subsegment_duration': subsegment_dur,
                    'starts_with_sap': st_sz >> 31,
                    'SAP_type': st_sz >> 28 & 7,
                    'SAP_delta_time': st_sz % 268435456
                })
        finally:
            fp.seek(self.start_of_box + self.size)


class SsixBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['subsegment_count'] = read_u32(fp)
            self.box_info['subsegment_list'] = []
            for i in range(self.box_info['subsegment_count']):
                subsegment_dict = {'range_count': read_u32(fp)}
                range_list = []
                for j in range(self.box_info['range_count']):
                    l_r = read_u32(fp)
                    range_list.append({'level': l_r // 16777216, 'range_size': l_r % 16777216})
                subsegment_dict['range_list'] = range_list
                self.box_info['subsegment_list'].append(subsegment_dict)
        finally:
            fp.seek(self.start_of_box + self.size)


class PrftBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['reference_track_id'] = read_u32(fp)
            self.box_info['ntp_timestamp'] = read_u64(fp)
            if self.version == 0:
                self.box_info['media_time'] = read_u32(fp)
            else:
                self.box_info['media_time'] = read_u64(fp)
        finally:
            fp.seek(self.start_of_box + self.size)





import numpy as np
import pandas as pd
import cv2
import pathlib
import math
import typing

from . import data_files, drawing, marker, ocv

class Coordinate(typing.NamedTuple):
    x: float = 0.
    y: float = 0.

class Plane:
    default_ref_image_name = 'reference_image.png'

    def __init__(self,
                 markers                : str|pathlib.Path|pd.DataFrame,                            # if str or Path: file from which to read markers. Else direction N_markerx4 array. Should contain centers of markers
                 marker_size            : float,                                                    # in "unit" units
                 plane_size             : Coordinate,                                               # in "unit" units

                 aruco_dict                                             = cv2.aruco.DICT_4X4_250,
                 marker_border_bits                                     = 1,
                 marker_pos_scale_fac                                   = 1.,                       # scale factor for marker positions in the markers input argument
                 unit                   : str                           = None,                     # Unit in which measurements (marker size and positions for instance) are expressed. Purely informational
                 package_to_read_from   : str                           = None,                     # if provided, reads marker file from specified package's resources
                 ref_image_store_path   : str|pathlib.Path              = None,
                 ref_image_size                                         = 1920                      # largest dimension
                 ):

        self.marker_size                                    = marker_size
        # marker positions
        self.markers            : dict[int,marker.Marker]   = {}
        self.plane_size                                     = plane_size
        self.bbox               : list[float]               = [0., 0., self.plane_size.x, self.plane_size.y]

        # marker specs
        self.aruco_dict                                     = cv2.aruco.getPredefinedDictionary(aruco_dict)
        self.marker_border_bits                             = marker_border_bits
        self.unit                                           = unit

        # prep markers
        self._load_markers(markers, marker_pos_scale_fac, package_to_read_from)

        # get reference image of plane
        if ref_image_store_path:
            ref_image_store_path = pathlib.Path(ref_image_store_path)

        # get image (always create reference image, to be safe (settings may have changed))
        img = self._store_reference_image(ref_image_store_path, ref_image_size)

        self._ref_image_size                                = ref_image_size
        self._ref_image_cache   : dict[int, np.ndarray]     = {ref_image_size: img}

    def set_origin(self, origin: Coordinate):
        # set origin of plane. Origin location is on current (not original) plane
        # so set_origin((5., 0.)) three times in a row shifts the origin rightward by 15 units
        for i in self.markers:
            self.markers[i].shift(-np.array(origin))

        self.bbox[0] -= origin.x
        self.bbox[2] -= origin.x
        self.bbox[1] -= origin.y
        self.bbox[3] -= origin.y

    def get_ref_image(self, im_size: int=None, as_RGB=False) -> np.ndarray:
        if im_size is None:
            im_size = self._ref_image_size
        if not isinstance(im_size,int):
            raise TypeError(f'width input should be an int, not {type(im_size)}')
        # check we have the image, if not, add to cache
        if im_size not in self._ref_image_cache:
            scale = float(im_size)/self._ref_image_size
            self._ref_image_cache[im_size] = cv2.resize(self._ref_image_cache[self._ref_image_size], None, fx=scale, fy=scale, interpolation = cv2.INTER_AREA)

        # return
        if as_RGB:
            return self._ref_image_cache[im_size][:,:,[2,1,0]]    # indexing returns a copy
        else:
            # OpenCV's BGR
            return self._ref_image_cache[im_size].copy()

    def draw(self, img, x, y, sub_pixel_fac=1, color=None, size=6):
        from . import transforms
        if not math.isnan(x):
            xy = transforms.to_image_pos(x, y, self.bbox, [img.shape[1], img.shape[0]])
            if color is None:
                drawing.openCVCircle(img, xy, 8, (0,255,0), -1, sub_pixel_fac)
                color = (0,0,0)
            drawing.openCVCircle(img, xy, size, color, -1, sub_pixel_fac)

    def _load_markers(self, markers: str|pathlib.Path|pd.DataFrame, marker_pos_scale_fac: float, package_to_read_from: str|None):
        # read in aruco marker positions
        markerHalfSizeMm  = self.marker_size/2.

        if isinstance(markers, pd.DataFrame):
            marker_pos = markers
        else:
            marker_pos = data_files.read_coord_file(markers, package_to_read_from)
        if marker_pos is None:
            raise RuntimeError(f"No markers could be read from the file {markers}, check it exists and contains markers")

        # turn into marker objects
        marker_pos.x *= marker_pos_scale_fac
        marker_pos.y *= marker_pos_scale_fac

        for idx, row in marker_pos.iterrows():
            c   = row[['x','y']].values
            # rotate markers (negative because plane coordinate system)
            rot = row[['rotation_angle']].values[0]
            rotr= -math.radians(rot)
            R   = np.array([[math.cos(rotr), math.sin(rotr)], [-math.sin(rotr), math.cos(rotr)]])
            # top left first, and clockwise: same order as detected ArUco marker corners
            tl = c + np.matmul(R, np.array([-markerHalfSizeMm, -markerHalfSizeMm]))
            tr = c + np.matmul(R, np.array([ markerHalfSizeMm, -markerHalfSizeMm]))
            br = c + np.matmul(R, np.array([ markerHalfSizeMm,  markerHalfSizeMm]))
            bl = c + np.matmul(R, np.array([-markerHalfSizeMm,  markerHalfSizeMm]))

            self.markers[idx] = marker.Marker(idx, c, corners=[tl, tr, br, bl], rot=rot)

    def get_aruco_board(self) -> cv2.aruco.Board:
        from . import aruco
        board_corner_points = []
        ids = []
        for key in self.markers:
            ids.append(key)
            marker_corner_points = np.vstack(self.markers[key].corners).astype('float32')
            board_corner_points.append(marker_corner_points)
        return aruco.create_board(board_corner_points, ids, self.aruco_dict)

    def _store_reference_image(self, path: pathlib.Path, im_size: int) -> np.ndarray:
        # get image with markers
        bbox_extents = [self.bbox[2]-self.bbox[0], math.fabs(self.bbox[3]-self.bbox[1])]  # math.fabs to deal with bboxes where (-,-) is bottom left
        aspect_ratio = bbox_extents[0]/bbox_extents[1]
        if aspect_ratio>1:
            width       = im_size
            height      = math.ceil(im_size/aspect_ratio)
        else:
            width       = math.ceil(im_size*aspect_ratio)
            height      = im_size

        img = np.zeros((height, width), np.uint8)
        img[:] = 255
        # collect all markers
        corner_points = []
        ids = []
        for key in self.markers:
            ids.append(key)
            corner_points.append(np.vstack(self.markers[key].corners).astype('float32'))

        # manually place the markers on the board
        # get info about marker positions on the board
        corner_points = np.dstack(corner_points)
        corner_points = corner_points-np.expand_dims(np.array([self.bbox[:2]]),2).astype('float32')

        # get position and size of marker in the generated image
        corner_points[:,0,:] = corner_points[:,0,:]/bbox_extents[0] * float(img.shape[1])
        corner_points[:,1,:] = corner_points[:,1,:]/bbox_extents[1] * float(img.shape[0])

        # get marker size
        pix_sz = np.vstack((np.hypot(corner_points[0,0,:]-corner_points[1,0,:], corner_points[0,1,:]-corner_points[1,1,:]),
                            np.hypot(corner_points[1,0,:]-corner_points[2,0,:], corner_points[1,1,:]-corner_points[2,1,:]))).T
        # marker should be square
        pix_sz = np.round(np.min(pix_sz,1)).astype('int')

        # place markers
        for i,sz,pos in zip(ids,pix_sz,np.moveaxis(corner_points, -1, 0)):
            # make marker
            marker_image = np.zeros((sz, sz), dtype=np.uint8)
            marker_image = self.aruco_dict.generateImageMarker(i, sz, marker_image, self.marker_border_bits)

            # put in image
            if pos[0,1]==pos[1,1] and pos[1,0]==pos[2,0] and pos[0,0]<pos[1,0]:
                # marker is aligned to image axes and not rotated, just blit
                ori = np.round(pos[0,:]).astype('int')
                img[ori[1]:ori[1]+sz, ori[0]:ori[0]+sz] = marker_image
                continue

            # set up affine transformation for placing marker in image
            in_corners = np.array([[-.5, -.5],
                                   [marker_image.shape[1]-.5, -.5],
                                   [marker_image.shape[1]-.5, marker_image.shape[0]-.5]]).astype('float32')
            transformation = cv2.getAffineTransform(in_corners, pos[:3,:])

            # perform affine transformation (i.e. rotate marker)
            img = cv2.warpAffine(marker_image, transformation, (img.shape[1], img.shape[0]), img, flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_TRANSPARENT)

        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        if path:
            cv2.imwrite(path, img)

        return img


class Pose:
    # description of tsv file used for storage
    _columns_compressed = {'frame_idx':1,
                           'pose_N_markers': 1, 'pose_reprojection_error': 1, 'pose_R_vec': 3, 'pose_T_vec': 3,
                           'homography_N_markers': 1, 'homography_mat': 9}
    _non_float          = {'frame_idx': int, 'pose_ok': bool, 'pose_N_markers': int, 'homography_N_markers': int}

    def __init__(self,
                 frame_idx              : int,
                 pose_N_markers         : int       = 0,
                 pose_reprojection_error: float     = -1.,
                 pose_R_vec             : np.ndarray= None,
                 pose_T_vec             : np.ndarray= None,
                 homography_N_markers   : int       = 0,
                 homography_mat         : np.ndarray= None):
        self.frame_idx              : int         = frame_idx
        # pose
        self.pose_N_markers         : int         = pose_N_markers        # number of ArUco markers this pose estimate is based on. 0 if failed
        self.pose_reprojection_error: float       = pose_reprojection_error
        self.pose_R_vec             : np.ndarray  = pose_R_vec
        self.pose_T_vec             : np.ndarray  = pose_T_vec
        # homography
        self.homography_N_markers   : int         = homography_N_markers  # number of ArUco markers this homongraphy estimate is based on. 0 if failed
        self.homography_mat         : np.ndarray  = homography_mat.reshape(3,3) if homography_mat is not None else homography_mat

        # internals
        self._RMat              = None
        self._RtMat             = None
        self._plane_normal      = None
        self._plane_point       = None
        self._RMatInv           = None
        self._RtMatInv          = None
        self._i_homography_mat  = None

    def pose_successful(self):
        return self.pose_N_markers>0
    def homography_successful(self):
        return self.homography_N_markers>0

    def draw_frame_axis(self, img, camera_params: ocv.CameraParams, arm_length, thickness, sub_pixel_fac, position = [0.,0.,0.]):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or not camera_params.has_intrinsics():
            return
        drawing.openCVFrameAxis(img, camera_params.camera_mtx, camera_params.distort_coeffs, self.pose_R_vec, self.pose_T_vec, arm_length, thickness, sub_pixel_fac, position)

    def cam_frame_to_world(self, point: np.ndarray):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMatInv is None:
            if self._RMatInv is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RMatInv = self._RMat.T
            self._RtMatInv = np.hstack((self._RMatInv,np.matmul(-self._RMatInv,self.pose_T_vec.reshape(3,1))))

        return np.matmul(self._RtMatInv,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def world_frame_to_cam(self, point: np.ndarray):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMat is None:
            if self._RMat is None:
                self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
            self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

        return np.matmul(self._RtMat,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def plane_to_cam_pose(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on plane (2D) to location on camera image (2D)
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)) or not camera_params.has_intrinsics():
            return np.full((2,), np.nan)
        return cv2.projectPoints(point, self.pose_R_vec, self.pose_T_vec, camera_params.camera_mtx, camera_params.distort_coeffs)[0].flatten()

    def cam_to_plane_pose(self, point: np.ndarray, camera_params: ocv.CameraParams) -> tuple[np.ndarray,np.ndarray]:
        # from location on camera image (2D) to location on plane (2D)
        # NB: also returns intermediate result (intersection with plane in camera space)
        from . import transforms
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)) or not camera_params.has_intrinsics():
            return np.full((2,), np.nan), np.full((3,), np.nan)

        g3D = transforms.unproject_point(*point, camera_params.camera_mtx, camera_params.distort_coeffs)

        # find intersection of 3D gaze with plane
        pos_cam = self.vector_intersect(g3D)  # default vec origin (0,0,0) because we use g3D from camera's view point

        # above intersection is in camera space, turn into world space to get position on plane
        (x,y,z) = self.cam_frame_to_world(pos_cam) # z should be very close to zero
        return np.asarray([x, y]), pos_cam

    def plane_to_cam_homography(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on plane (2D) to location on camera image (2D)
        from . import transforms
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if self._i_homography_mat is None:
            self._i_homography_mat = np.linalg.inv(self.homography_mat)
        out = transforms.apply_homography(self._i_homography_mat, *point)
        if camera_params.has_intrinsics():
            out = transforms.distort_point(*out, camera_params.camera_mtx, camera_params.distort_coeffs)
        return out

    def cam_to_plane_homography(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on camera image (2D) to location on plane (2D)
        from . import transforms
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if camera_params.has_intrinsics():
            point = transforms.undistort_point(*point, camera_params.camera_mtx, camera_params.distort_coeffs)
        return transforms.apply_homography(self.homography_mat, *point)

    def get_origin_on_image(self, camera_params: ocv.CameraParams) -> np.ndarray:
        if self.pose_successful() and camera_params.has_intrinsics():
            a = self.plane_to_cam_pose(np.zeros((1,3)), camera_params)
        elif self.homography_successful():
            a = self.plane_to_cam_homography([0., 0.], camera_params)
        else:
            a = np.full((2,), np.nan)
        return a

    def vector_intersect(self, vector: np.ndarray, origin = np.array([0.,0.,0.])):
        from . import transforms

        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(vector)):
            return np.full((3,), np.nan)

        if self._plane_normal is None:
            if self._RtMat is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

            # get poster normal
            self._plane_normal = self._RMat[:,2]     # equivalent to: np.matmul(self._RMat, np.array([0., 0., 1.]))
            # get point on poster (just use origin)
            self._plane_point  = self._RtMat[:,3]    # equivalent to: np.matmul(self._RtMat, np.array([0., 0., 0., 1.]))

        # normalize vector
        vector /= np.linalg.norm(vector)

        # find intersection of 3D gaze with poster
        return transforms.intersect_plane_ray(self._plane_normal, self._plane_point, vector.flatten(), origin.flatten())


def read_dict_from_file(fileName:str|pathlib.Path, episodes:list[list[int]]=None) -> dict[int,Pose]:
    return data_files.read_file(fileName,
                                Pose, True, True, False, False,
                                episodes=episodes)[0]

def write_list_to_file(poses: list[Pose], fileName:str|pathlib.Path, skip_failed=False):
    data_files.write_array_to_file(poses, fileName,
                                    Pose._columns_compressed,
                                    skip_all_nan=skip_failed)